from __future__ import annotations

import logging
import re
from typing import TYPE_CHECKING

from dateutil.relativedelta import relativedelta
from django.forms import ValidationError
from django.utils import timezone

from firefighter.incidents.views.date_utils import (
    get_bounds_from_calendar_year,
    get_bounds_from_quarter,
    get_bounds_from_week,
    get_bounds_from_year,
    get_current_quarter,
    get_last_quarter,
    parse_date_natural,
)

if TYPE_CHECKING:
    from datetime import datetime

logger = logging.getLogger(__name__)


RANGE_REGEX = re.compile(
    r"^(?P<iso_year>[iI])?(?P<year>\d{4})?/?(?:[qQ](?P<quarter>[1-4])|([wW](?P<week>5[0-3]|[1-4]\d|[0-4]?[1-9])))?$"
)


SPECIAL_RANGES = ["current_week", "last_week", "current_quarter", "last_quarter"]


def parse_moment(value: str) -> slice | None:
    """Returns a :class:`slice` with the :func:`slice.start` and :func:`slice.stop` of the given moment.
    If the moment is not a valid moment, returns None.
    If the moment is a point in time, returns a slice with the same start and stop.
    """
    value = value.strip()

    # Quarter/Week relative (current/previous quarter, current/previous week)
    if value in SPECIAL_RANGES:
        now = timezone.localtime(timezone.now())
        returned_range = get_date_range_from_relative_calendar_value(value, now, value)
        if None not in returned_range:
            return slice(returned_range[0], returned_range[1])

    # Match according to the regex (W20, Q4, YYYY-W20, YYYY-Q4...)
    week_or_quarter_time_frame = (*get_date_range_from_calendar_value(value), value)
    if None not in week_or_quarter_time_frame:
        return slice(week_or_quarter_time_frame[0], week_or_quarter_time_frame[1])

    # Match using dateparser (lots of formats)
    return parse_date_natural(value)


def get_date_range_from_special_date(
    unparsed_date: str,
) -> tuple[datetime | None, datetime | None, str | None, str | None]:
    """TODO Specify the year parameter (ISO or calendar)."""
    if "-" in unparsed_date:
        logger.debug("Parsing range, splitting with /")

        if unparsed_date.count("-") == 1:
            split = unparsed_date.split("-")
        elif unparsed_date.count(" - ") == 1:
            split = unparsed_date.split(" - ")
        else:
            err_msg = f"Invalid range: {unparsed_date}. Only one separator is allowed"
            raise ValidationError(err_msg)
        if len(split) == 2:
            for i in range(2):
                split[i] = split[i].strip()

            beg = parse_moment(split[0])
            end = parse_moment(split[1])
            logger.debug(end)
            return (
                beg.start if beg else None,
                end.stop if end else None,
                None,
                split[0] + "-" + split[1],
            )
        # TODO Message warning + logger.warning
        raise ValidationError(
            "Invalid range: %(unparsed_date)s. Only one separator is allowed",
            params={"unparsed_date": unparsed_date},
            code="invalid",
        )
    try:
        date_range = parse_moment(unparsed_date)
    except ValueError as exc:
        raise ValidationError(
            "Invalid range: %(unparsed_date)s. %(error)s",
            params={"unparsed_date": unparsed_date, "error": str(exc)},
            code="invalid",
        ) from exc
    if date_range:
        return date_range.start, date_range.stop, None, "TODO"

    # TODO Message warning + logger.warning
    return None, None, None, None


def get_date_range_from_relative_calendar_value(
    unparsed_date: str, now: datetime, date_input: str
) -> tuple[datetime, datetime, str, str] | tuple[None, None, None, None]:
    if unparsed_date == "current_week":
        year, week, _weekday = now.date().isocalendar()
        return *get_bounds_from_week(year, week), date_input
    if unparsed_date == "last_week":
        year, week, _weekday = (now.date() + relativedelta(days=-7)).isocalendar()
        return *get_bounds_from_week(year, week), date_input
    if unparsed_date == "current_quarter":
        quarter_year, quarter_number = get_current_quarter()
        return *get_bounds_from_quarter(quarter_year, quarter_number), date_input
    if unparsed_date == "last_quarter":
        quarter_year, quarter_number = get_last_quarter()
        return *get_bounds_from_quarter(quarter_year, quarter_number), date_input
    return None, None, None, None


def get_date_range_from_calendar_value(
    date_range: str,
) -> tuple[datetime | None, datetime | None, str | None]:
    match = RANGE_REGEX.match(date_range)

    if not match:
        return None, None, None

    dict_match = match.groupdict()

    year = (
        int(dict_match["year"])
        if dict_match.get("year")
        else timezone.localtime(timezone.now()).date().year
    )

    if dict_match.get("quarter"):
        return get_bounds_from_quarter(year, int(dict_match["quarter"]))
    if dict_match.get("week"):
        return get_bounds_from_week(year, int(dict_match["week"]))
    if dict_match.get("year"):
        if dict_match.get("iso_year"):
            return get_bounds_from_year(year)
        return get_bounds_from_calendar_year(year)

    return None, None, None


def get_range_look_args(
    gte: datetime | None, lte: datetime | None, field_name: str
) -> dict[str, datetime]:
    if gte and lte:
        return {
            f"{field_name}__gte": gte,
            f"{field_name}__lte": lte,
        }
    if lte:
        return {f"{field_name}__lte": lte}
    if gte:
        return {
            f"{field_name}__gte": gte,
        }
    return {}



from __future__ import annotations

import logging
from datetime import UTC, datetime
from typing import TYPE_CHECKING, Any

from django.db.utils import IntegrityError
from slack_sdk.errors import SlackApiError

from firefighter.firefighter.utils import get_in
from firefighter.slack.models.message import Message
from firefighter.slack.models.user import SlackUser
from firefighter.slack.slack_app import SlackApp
from firefighter.slack.slack_incident_context import get_incident_from_context

if TYPE_CHECKING:
    from slack_sdk.web.client import WebClient

app = SlackApp()

logger = logging.getLogger(__name__)


def reaction_on_message_only(event: dict[str, Any]) -> bool:
    return bool(get_in(event, ["item", "type"]) == "message")


def special_reaction_only(event: dict[str, Any]) -> bool:
    return event.get("reaction") == "mega"


@app.event("reaction_added", matchers=[reaction_on_message_only, special_reaction_only])
def reaction_added(event: dict[str, Any], client: WebClient) -> None:
    logger.debug(event)

    # Check that the event is a message reaction with the right emoji
    item = event.get("item")
    if not isinstance(item, dict):
        err_msg = f"Event item is not a dict: {item}"
        raise TypeError(err_msg)

    # Get needed variables and assert they are here
    conversation_id = item.get("channel")
    message_ts = item.get("ts")
    if not isinstance(message_ts, str | float):
        err_msg = f"Event message_ts is not a str or float: {message_ts}"
        raise TypeError(err_msg)

    # The timestamp is UTC
    message_ts_dt = datetime.fromtimestamp(float(message_ts), tz=UTC)
    user_reaction = event.get("user")

    if not isinstance(conversation_id, str):
        err_msg = f"Event conversation_id is not a str: {conversation_id}"
        raise TypeError(err_msg)
    if not isinstance(user_reaction, str):
        err_msg = f"Event user_reaction is not a str: {user_reaction}"
        raise TypeError(err_msg)
    if not isinstance(message_ts, str):
        err_msg = f"Event message_ts is not a str: {message_ts_dt}"
        raise TypeError(err_msg)

    # Check that the message is in an incident conversation
    incident = get_incident_from_context(event)
    if not incident:
        return

    user = SlackUser.objects.get_user_by_slack_id(slack_id=user_reaction)
    if not user or not hasattr(user, "slack_user") or not user.slack_user:
        logger.error("User %s could not be found.", user_reaction)
        return

    # Get the message as it is not included in the event
    try:
        result = client.conversations_history(
            channel=conversation_id, inclusive=True, oldest=message_ts, limit=1
        )

    except SlackApiError:
        logger.exception("Could not get the message from reaction!")
        return

    messages = result.get("messages")
    if not isinstance(messages, list):
        err_msg = f"Event messages is not a list: {messages}"
        raise TypeError(err_msg)

    message = messages[0]
    message_text = message.get("text")
    logger.debug(message)
    if not isinstance(message_text, str):
        err_msg = f"Event message_text is not a str: {message_text}"
        raise TypeError(err_msg)

    try:
        message = Message(
            ts=message_ts_dt,
            text=message_text,
            type=message["type"],
            blocks=message.get("blocks"),
            conversation_id=incident.conversation.id,
            user_id=user.slack_user.id,
        )
        message.save()
    except IntegrityError:
        # We do not post in a thread as the client won't show it.
        client.chat_postEphemeral(
            channel=conversation_id,
            user=user_reaction,
            text=f"<@{user_reaction}> :x: An Incident Update has already been created from this message!",
        )
        return

    incident_update = incident.create_incident_update(
        message=message_text,
        created_by=user,
        event_ts=message_ts_dt,
    )
    message.incident_update = incident_update
    message.save()


@app.event("reaction_added")
def reaction_added_ignore() -> None:
    """Ignore other reaction_added events.
    Must be the last event handler for reaction_added.
    """
    return



from __future__ import annotations

import logging
import textwrap
from typing import TYPE_CHECKING, Any

from django.conf import settings
from django.db import models
from slack_sdk.errors import SlackApiError

from firefighter.incidents.models.incident import Incident
from firefighter.slack.models.conversation import Conversation, ConversationManager
from firefighter.slack.models.user import SlackUser
from firefighter.slack.slack_app import DefaultWebClient, slack_client
from firefighter.slack.utils import channel_name_from_incident

if TYPE_CHECKING:
    from slack_sdk.web.client import WebClient
    from slack_sdk.web.slack_response import SlackResponse

    from firefighter.incidents.models import User

APP_DISPLAY_NAME: str = settings.APP_DISPLAY_NAME
SLACK_APP_EMOJI: str = settings.SLACK_APP_EMOJI
logger = logging.getLogger(__name__)


class IncidentChannelManager(ConversationManager["IncidentChannel"]):
    @slack_client
    def create_incident_channel(
        self,
        incident: Incident,
        client: WebClient = DefaultWebClient,
        **kwargs: Any,
    ) -> IncidentChannel | None:
        channel_name = channel_name_from_incident(incident)

        logger.debug("Creating channel #%s", channel_name)
        channel: IncidentChannel | None = super().create_channel(
            name=channel_name,
            client=client,
            incident=incident,
            is_private=incident.private,
            **kwargs,
        )
        return channel


class IncidentChannel(Conversation):
    objects: IncidentChannelManager = IncidentChannelManager()  # type: ignore[assignment]

    incident = models.OneToOneField[Incident, Incident](
        Incident,
        on_delete=models.CASCADE,
        related_name="conversation",
    )
    if TYPE_CHECKING:
        incident_id: int

    @slack_client
    def rename_if_needed(self, client: WebClient = DefaultWebClient) -> bool | None:
        new_name = self.channel_name_from_incident()

        if self.name != new_name:
            try:
                slack_conv = client.conversations_rename(
                    channel=self.channel_id, name=new_name
                )
            except SlackApiError:
                logger.exception(f"Could not rename channel {self.id}.")
                return None

            (
                channel_id,
                channel_name,
                channel_type,
                status,
            ) = Conversation.objects.parse_slack_response(slack_conv=slack_conv)

            self.name = channel_name
            self.status = status
            self.channel_id = channel_id
            self.type = channel_type
            self.save()
            return True
        return None

    def channel_name_from_incident(self, incident: Incident | None = None) -> str:
        if incident and self.incident != incident:
            raise ValueError(
                "The incident passed as argument does not match the incident of the channel."
            )
        if incident is None:
            incident = self.incident

        return channel_name_from_incident(incident)

    @slack_client
    def set_incident_channel_topic(
        self, client: WebClient = DefaultWebClient
    ) -> SlackResponse | None:
        incident: Incident = self.incident
        topic = f"Incident - {incident.priority.emoji} {incident.priority.name}  - {incident.status.label} - {incident.component.group.name} - {incident.component.name} - {SLACK_APP_EMOJI} <{incident.status_page_url + '?utm_medium=FireFighter+Slack&utm_source=Slack+Topic&utm_campaign=Slack+Topic+Link'}| {APP_DISPLAY_NAME} Status>"

        if len(topic) > 250:
            logger.warning(
                f"Generated a Slack topic longer than the maximum! It will be truncated to 250 chars. Topic: {topic}"
            )
            topic = textwrap.shorten(topic, width=250, placeholder="...")

        return self.update_topic(topic, client=client)

    @slack_client
    def invite_users(
        self, users_mapped: list[User], client: WebClient = DefaultWebClient
    ) -> None:
        """Invite users to the conversation, if they have a Slack user linked and are active.

        Try to invite all users as batch, but if some fail, continue individually.
        """
        users_with_slack: list[User] = self._get_active_slack_users(users_mapped)
        users_with_slack = list(set(users_with_slack))  # Remove duplicates

        user_id_list: set[str] = self._get_slack_id_list(users_with_slack)
        if not user_id_list:
            logger.info(f"No users to invite to the conversation {self}.")
            return

        logger.info(f"Inviting users with SlackIDs: {user_id_list}")

        invited_slack_user_ids = self._invite_users_to_conversation(
            user_id_list, client
        )
        invited_users = {
            u
            for u in users_with_slack
            if u.slack_user and u.slack_user.slack_id in invited_slack_user_ids
        }

        if invited_slack_user_ids != set(user_id_list):
            logger.warning(
                f"Could not invite all users to the conversation {self}. Missing users: {user_id_list - invited_slack_user_ids}"
            )

        self.incident.members.add(*invited_users)

    @staticmethod
    def _get_active_slack_users(users_mapped: list[User]) -> list[User]:
        """Filter out users that have no Slack user linked or are disabled."""
        users_with_slack: list[User] = []

        for user in users_mapped:
            user_with_slack = SlackUser.objects.add_slack_id_to_user(user)
            if user_with_slack is not None and user_with_slack.is_active:
                users_with_slack.append(user_with_slack)
            else:
                logger.warning(
                    f"User {user.id} has no Slack user linked or is disabled."
                )

        return users_with_slack

    @staticmethod
    def _get_slack_id_list(users_with_slack: list[User]) -> set[str]:
        return {u.slack_user.slack_id for u in users_with_slack if u.slack_user}

    def _invite_users_to_conversation(
        self, user_id_list: set[str], client: WebClient
    ) -> set[str]:
        try:
            done = self._invite_users_with_slack_id(user_id_list, client)
        except SlackApiError:
            logger.warning(
                f"Could not batch import Slack users! Slack IDs: {user_id_list}",
                exc_info=True,
            )
            done = self._invite_users_to_conversation_individually(user_id_list, client)
        return done

    def _invite_users_with_slack_id(
        self, user_id_list: set[str], client: WebClient
    ) -> set[str]:
        client.conversations_invite(
            channel=self.channel_id,
            users=list(user_id_list),
        )

        logger.info(
            f"Imported Slack IDs: {user_id_list} in channel {self.channel_id} for incident {self.incident.id}"
        )
        return user_id_list

    def _invite_users_to_conversation_individually(
        self, slack_user_ids: set[str], client: WebClient
    ) -> set[str]:
        done = set()
        for slack_user_id in slack_user_ids:
            try:
                self._invite_users_with_slack_id({slack_user_id}, client)
                done.add(slack_user_id)
            except SlackApiError:
                logger.warning(
                    f"Could not import Slack user! User ID: {slack_user_id}",
                    exc_info=True,
                )
        return done


class BatchInviteError(SlackApiError):
    pass



# Generated by Django 4.2.7 on 2023-12-04 10:46

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    replaces = [
        ("raid", "0001_initial_squashed_0008_qualifierrotation"),
        ("raid", "0009_alter_jiraticket_business_impact"),
        ("raid", "0010_raidarea"),
        ("raid", "0010_jiraticket_issue_type_jiraticket_project_key"),
        ("raid", "0011_merge_20230408_1956"),
        ("raid", "0012_jiraticketimpact_jiraticket_impacts"),
        ("raid", "0013_alter_jiraticket_watchers"),
        ("raid", "0014_alter_jiraticket_business_impact"),
        ("raid", "0015_alter_jiraticket_assignee_alter_jiraticket_reporter_and_more"),
        ("raid", "0016_remove_jiraticket_assignee_and_more"),
        ("raid", "0017_alter_qualifierrotation_options_and_more"),
        ("raid", "0018_alter_qualifier_jira_user_and_more"),
    ]

    dependencies = [
        ("jira_app", "0005_alter_jiraissue_assignee"),
        ("incidents", "0036_severity_enabled_create_severity_enabled_update_and_more"),
        (
            "incidents",
            "0039_impact_impacttype_incidentimpact_impact_impact_type_and_more",
        ),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("jira_app", "0001_initial"),
        ("jira_app", "0003_jiraissue"),
    ]
    operations = [
        migrations.CreateModel(
            name="JiraTicket",
            fields=[
                (
                    "jiraissue_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="jira_app.jiraissue",
                    ),
                ),
                (
                    "business_impact",
                    models.CharField(
                        blank=True,
                        default="",
                        help_text="It maps with customfield_10936",
                        max_length=128,
                        null=True,
                    ),
                ),
            ],
            options={
                "verbose_name": "Jira ticket",
                "verbose_name_plural": "Jira tickets",
            },
            bases=("jira_app.jiraissue",),
        ),
        migrations.CreateModel(
            name="RaidArea",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                ("name", models.CharField(max_length=128)),
                (
                    "area",
                    models.CharField(
                        choices=[
                            ("Sellers", "Sellers"),
                            ("Internal", "Internal"),
                            ("Customers", "Customers"),
                        ],
                        max_length=32,
                    ),
                ),
            ],
            options={
                "unique_together": {("name", "area")},
            },
        ),
        migrations.CreateModel(
            name="QualifierRotation",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                ("day", models.DateField(unique=True)),
                (
                    "protected",
                    models.BooleanField(
                        default=False,
                        help_text="If checked, it is because this rotation was previously modified so this date won't be deleted and it can not be unprotected again from here.",
                    ),
                ),
                (
                    "jira_user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="jira_app.jirauser",
                    ),
                ),
            ],
            options={
                "verbose_name": "Qualifiers rotation",
                "verbose_name_plural": "Qualifiers rotation",
            },
        ),
        migrations.CreateModel(
            name="Qualifier",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "jira_user",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="jira_app.jirauser",
                    ),
                ),
            ],
            options={
                "verbose_name": "Qualifier",
                "verbose_name_plural": "Qualifiers",
            },
        ),
        migrations.CreateModel(
            name="JiraTicketImpact",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "impact",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="incidents.impact",
                    ),
                ),
                (
                    "jira_ticket",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="raid.jiraticket",
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="jiraticket",
            name="impacts",
            field=models.ManyToManyField(
                through="raid.JiraTicketImpact", to="incidents.impact"
            ),
        ),
        migrations.AddField(
            model_name="jiraticket",
            name="incident",
            field=models.OneToOneField(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="jira_ticket",
                to="incidents.incident",
            ),
        ),
    ]



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, cast

from django.conf import settings
from django.contrib import messages
from django.contrib.auth.mixins import LoginRequiredMixin
from django.urls import reverse_lazy
from django.views import generic

from firefighter.pagerduty.forms.create_pagerduty_incident import (
    CreatePagerDutyIncidentFreeForm,
)
from firefighter.pagerduty.tasks.trigger_oncall import trigger_oncall
from firefighter.slack.models.conversation import Conversation

if TYPE_CHECKING:
    from django.http import HttpResponse

    from firefighter.firefighter.utils import HtmxHttpRequest
logger = logging.getLogger(__name__)


class CreatePagerDutyIncidentFreeView(
    LoginRequiredMixin, generic.FormView[CreatePagerDutyIncidentFreeForm]
):
    form_class = CreatePagerDutyIncidentFreeForm
    success_url = reverse_lazy("pagerduty:oncall-list")

    def get_template_names(self) -> list[str]:
        request = cast("HtmxHttpRequest", self.request)
        if request.htmx and not request.htmx.boosted:
            template_name = "partials/trigger_oncall_form_view_modal.html"
        else:
            template_name = "pages/oncall_trigger.html"

        return [template_name]

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        context = super().get_context_data(**kwargs)
        context["page_title"] = "Trigger on-call"
        return context

    def form_valid(self, form: CreatePagerDutyIncidentFreeForm) -> HttpResponse:
        incident_key = form.cleaned_data["title"].replace(" ", "_").lower()[0:255]
        conference = Conversation.objects.get_or_none(tag="tech_incidents")
        conference_url = settings.BASE_URL if conference is None else conference.link
        try:
            # XXX Maybe do more one day with the response!
            trigger_oncall(
                title=form.cleaned_data["title"],
                oncall_service=form.cleaned_data["service"],
                details=form.cleaned_data["details"],
                triggered_by=self.request.user,  # type: ignore
                incident_key=incident_key,
                conference_url=conference_url,
            )
        except Exception:  # XXX Do not catch blindly
            logger.exception("On-call trigger failed!")
            messages.error(self.request, "On-call trigger failed!")
            return super().form_valid(form)
        messages.success(self.request, "On-call triggered")
        return super().form_valid(form)



from __future__ import annotations

import logging

from django.apps import AppConfig

logger = logging.getLogger(__name__)


class FireFighter(AppConfig):
    name = "firefighter.firefighter"
    label = "firefighter"
    verbose_name = "FireFighter"

    def ready(self) -> None:
        from django.conf import settings

        import firefighter.components

        if settings.FF_SKIP_SECRET_KEY_CHECK:
            logger.info("Skipping SECRET_KEY check.")
            return

        env: str = settings.ENV
        secret_key: str = settings.SECRET_KEY

        if _entropy(secret_key) < 4:
            if env == "dev":
                logger.warning(
                    "SECRET_KEY is too weak, please change it to a stronger one. This is a warning for ENV=dev, but will raise an error in other environments."
                )
            else:
                raise ValueError(
                    "SECRET_KEY is too weak, please change it to a stronger one."
                )


def _entropy(value: str) -> float:
    """Calculate the entropy of a string."""
    import math
    import string

    if not value:
        return 0.0
    entropy: float = 0.0
    for x in string.printable:
        p_x = float(value.count(x)) / len(value)
        if p_x > 0:
            entropy += -p_x * math.log2(p_x)
    return entropy



from __future__ import annotations

from typing import Any
from unittest.mock import MagicMock

import pytest
from slack_sdk.models.blocks.block_elements import ButtonElement
from slack_sdk.models.blocks.blocks import (
    ActionsBlock,
)

from firefighter.incidents.forms.create_incident import CreateIncidentFormBase
from firefighter.incidents.models.user import User
from firefighter.slack.views.modals.open import OpeningData, OpenModal
from firefighter.slack.views.modals.opening.set_details import SetIncidentDetails


def build_opening_data(**kwargs: Any) -> OpeningData:
    data: OpeningData = {
        "response_type": kwargs.get("response_type", "normal"),
        "impact_form_data": kwargs.get("impact_form_data", {}),
        "details_form_data": kwargs.get("details_form_data", {}),
        "incident_type": kwargs.get("incident_type"),
    }
    return data


@pytest.fixture
def open_incident_context() -> OpeningData:
    return build_opening_data()


@pytest.fixture
def user() -> User:
    return User(username="testuser", email="testuser@example.com")


@pytest.fixture
def ack() -> MagicMock:
    return MagicMock()


def test_get_done_review_blocks_can_submit_false(
    open_incident_context: OpeningData, user: User
) -> None:
    blocks = OpenModal.get_done_review_blocks(
        open_incident_context,
        user,
        details_form_done=False,
        details_form_class=None,
        details_form=None,
        can_submit=False,
    )
    assert len(blocks) == 0


@pytest.mark.django_db
def test_check_impact_form_invalid(open_incident_context: OpeningData) -> None:
    open_incident_context["impact_form_data"] = {"impact": "invalid"}  # type: ignore[dict-item]
    result = OpenModal._check_impact_form(open_incident_context)
    assert result is False


def test_validate_details_form_valid() -> None:
    details_form_modal_class = MagicMock(spec=SetIncidentDetails)
    details_form_class = MagicMock(spec=CreateIncidentFormBase)
    details_form_data = {"key": "value"}

    details_form_modal_class.form_class = details_form_class

    details_form_instance = MagicMock(spec=CreateIncidentFormBase)
    details_form_instance.is_valid.return_value = True
    details_form_class.return_value = details_form_instance

    is_valid, returned_form_class, returned_form = OpenModal._validate_details_form(
        details_form_modal_class, details_form_data
    )

    assert is_valid is True
    assert returned_form_class == details_form_class
    assert returned_form == details_form_instance


def test_validate_details_form_invalid() -> None:
    details_form_modal_class = MagicMock(spec=SetIncidentDetails)
    details_form_class = MagicMock(spec=CreateIncidentFormBase)
    details_form_data = {"key": "value"}

    details_form_modal_class.form_class = details_form_class

    details_form_instance = MagicMock(spec=CreateIncidentFormBase)
    details_form_instance.is_valid.return_value = False
    details_form_class.return_value = details_form_instance

    is_valid, returned_form_class, returned_form = OpenModal._validate_details_form(
        details_form_modal_class, details_form_data
    )

    assert is_valid is False
    assert returned_form_class == details_form_class
    assert returned_form == details_form_instance


def test_build_response_type_blocks_bis(open_incident_context: OpeningData) -> None:
    open_incident_context["response_type"] = "critical"
    blocks = OpenModal._build_response_type_blocks(open_incident_context)

    assert len(blocks) == 1
    first_block = blocks[0]
    assert isinstance(first_block, ActionsBlock)
    assert len(first_block.elements) == 1
    assert all(isinstance(element, ButtonElement) for element in first_block.elements)


@pytest.mark.django_db
def test_build_modal_fn_empty(user: User) -> None:
    open_incident_context = OpeningData()
    view = OpenModal().build_modal_fn(open_incident_context, user=user)
    assert view.type == "modal"
    # View should only have 3 blocks
    assert len(view.blocks) == 3

    # View should not have a submit button
    assert view.submit is None



from __future__ import annotations

import logging
import re
from functools import cache
from typing import TYPE_CHECKING, Final, TypeVar

if TYPE_CHECKING:
    from firefighter.jira_app.types import (
        StatusTransitionInfo,
        _StatusId,
        _StatusName,
        _TargetStatusId,
        _TransitionName,
    )

_SNAKE_CASE_PATTERN: Final[re.Pattern[str]] = re.compile(r"(.)([A-Z]+)")

T = TypeVar("T")

logger = logging.getLogger(__name__)


def _transition_path(
    statuses_transitions: dict[T, set[T]], from_state: T, to_state: T
) -> list[T]:
    """Returns the path to transition from one status to another.
        Includes the from_state and to_state.

        If the path is not found, returns an empty list.

    Args:
        statuses_transitions (dict[T, set[T]]): dict of statuses and their targets
        from_state (T): starting status
        to_state (T): ending status

    Returns:
        list[T]: list of statuses to transition from from_state to to_state
    """
    if from_state == to_state:
        return []
    visited = set()
    queue = [[from_state]]
    while queue:
        path = queue.pop(0)
        node = path[-1]
        if node not in visited:
            visited.add(node)
            for adjacent in statuses_transitions.get(node, []):
                new_path = list(path)
                new_path.append(adjacent)
                queue.append(new_path)
                if adjacent == to_state:
                    return new_path
    logger.warning(f"No path found for from_state {from_state} to to_state {to_state}")
    return []


def _states_to_transitions_names(
    statuses_transitions_names: dict[_StatusId, dict[_StatusId, _StatusName]],
    states: list[_StatusId],
) -> list[_StatusName]:
    return [
        statuses_transitions_names[states[i]][states[i + 1]]
        for i in range(len(states) - 1)
    ]


def get_transitions_to_apply(
    current_status_id: _StatusId,
    transitions_info: list[StatusTransitionInfo],
    closed_state_id: _StatusId,
) -> list[_StatusName]:
    # Get path to closed state
    statuses_targets: dict[_StatusId, set[_TargetStatusId]] = {
        k["status_id"]: k["target_statuses"] for k in transitions_info
    }
    path_states = _transition_path(statuses_targets, current_status_id, closed_state_id)

    # Get transitions to apply from path
    statuses_transitions_names: dict[
        _StatusId, dict[_TargetStatusId, _TransitionName]
    ] = {k["status_id"]: k["transition_to_status"] for k in transitions_info}
    return _states_to_transitions_names(statuses_transitions_names, path_states)


def get_status_id_from_name(
    transitions_info: list[StatusTransitionInfo], status_name: str
) -> _StatusId | None:
    return next(
        (k["status_id"] for k in transitions_info if k["status_name"] == status_name),
        None,
    )


@cache
def _snake_case_key(key: str) -> str:
    s1 = _SNAKE_CASE_PATTERN.sub(r"\1_\2", key)
    return s1.lower()


def pythonic_keys(d: T) -> T:
    """Converts camelCase keys in a dict or list of dict to snake_case. Works recursively."""
    if isinstance(d, dict):
        new_d = {}
        for key, value in d.items():
            new_key = _snake_case_key(key)
            new_d[new_key] = pythonic_keys(value)
        return new_d  # type: ignore[return-value]
    if isinstance(d, list):
        return [pythonic_keys(v) for v in d]  # type: ignore[return-value]
    return d



from __future__ import annotations

from firefighter.firefighter.settings.components.common import INSTALLED_APPS
from firefighter.firefighter.settings.settings_utils import config

ENABLE_CONFLUENCE = config("ENABLE_CONFLUENCE", cast=bool, default=False)
"Enable the Confluence app."

if ENABLE_CONFLUENCE:
    INSTALLED_APPS += ("firefighter.confluence",)

    CONFLUENCE_RUNBOOKS_FOLDER_ID: int = config(
        "CONFLUENCE_RUNBOOKS_FOLDER_ID", cast=int
    )
    "The Confluence page ID where runbooks are stored."

    CONFLUENCE_POSTMORTEM_TEMPLATE_PAGE_ID: int = config(
        "CONFLUENCE_POSTMORTEM_TEMPLATE_PAGE_ID", cast=int
    )
    "The Confluence page ID of the template to use for postmortems."

    CONFLUENCE_POSTMORTEM_FOLDER_ID: int = config(
        "CONFLUENCE_POSTMORTEM_FOLDER_ID", cast=int
    )
    "The Confluence page ID  where to create and nest postmortems."

    CONFLUENCE_POSTMORTEM_SPACE: str = config("CONFLUENCE_POSTMORTEM_SPACE")
    "XXX To rename CONFLUENCE_DEFAULT_SPACE. The Confluence space where to create pages by default, mainly for postmortems."

    CONFLUENCE_ON_CALL_PAGE_ID: int | None = config(
        "CONFLUENCE_ON_CALL_PAGE_ID", cast=int, default=None
    )
    "The Confluence page ID where to export the current on-call schedule. If not set, export tasks will be skipped."

    CONFLUENCE_USERNAME: str = config("CONFLUENCE_USERNAME")
    "The Confluence username to use."
    CONFLUENCE_API_KEY: str = config("CONFLUENCE_API_KEY")
    "The Confluence API key to use."
    CONFLUENCE_URL: str = config("CONFLUENCE_URL")
    "The Confluence URL to use. If no protocol is defined, https will be used."
    CONFLUENCE_MOCK_CREATE_POSTMORTEM: bool = config(
        "CONFLUENCE_MOCK_CREATE_POSTMORTEM", default=False, cast=bool
    )



from __future__ import annotations

import uuid

from django.db import models
from django_stubs_ext.db.models import TypedModelMeta

from firefighter.slack.models.conversation import Conversation
from firefighter.slack.models.user_group import UserGroup


class Sos(models.Model):
    """A SOS is a target with a name, a conversation, and optionally a user group.

    Incident responders can use it with `/incident sos` to ask for help to a specific group of people.

    Helpers will be notified in the selected conversation, and the user group will be mentioned (or `@here` if no user group is selected)
    """

    # XXX We need a relation to `Gear` or `Group` here, but current data is not good enough.
    id = models.UUIDField(primary_key=True, editable=False, default=uuid.uuid4)
    name = models.CharField(max_length=255)
    user_group = models.ForeignKey(
        UserGroup, blank=True, null=True, on_delete=models.CASCADE
    )
    conversation = models.ForeignKey(
        Conversation, blank=True, on_delete=models.CASCADE
    )

    class Meta(TypedModelMeta):
        verbose_name = "SOS"
        verbose_name_plural = "SOS"

    def __str__(self) -> str:
        return self.name

    @property
    def usergroup_slack_fmt(self) -> str:
        """Returns either `@usergroup` or `@here` depending on usergroup presence."""
        return f"@{self.user_group.handle}" if self.user_group else "@here"



from __future__ import annotations

BASE_TABLE_ATTRS: dict[str, dict[str, str] | str] = {
    "class": "table-auto divide-y divide-neutral-200 dark:divide-neutral-600 w-full rounded-md",
    "th": {
        "class": "px-2 py-3 tracking-wider hover:underline uppercase text-xs font-medium  whitespace-nowrap",
        "scope": "col",
    },
    "td": {"class": "px-3 py-4 whitespace-nowrap  text-center"},
    "thead": {
        "class": "bg-neutral-50 dark:bg-neutral-700  text-neutral-500 dark:text-neutral-50 rounded-t-md",
        "scope": "col",
    },
    "tbody": {
        "class": "bg-base-100 dark:bg-neutral-800 divide-y divide-neutral-200 text-base-content dark:divide-neutral-600 text-sm font-medium"
    },
}
# XXX HTMX support for pagination/django-tables2 ordering
# XXX Reuse more columns and styles!



from __future__ import annotations

import logging
import re
from typing import TYPE_CHECKING, Any, Final

from django.utils.timezone import localtime
from slack_sdk.errors import SlackApiError

from firefighter.slack.slack_app import DefaultWebClient, slack_client

if TYPE_CHECKING:
    from collections.abc import Sequence

    from slack_sdk.models.blocks import Block
    from slack_sdk.web.client import WebClient

    from firefighter.incidents.models.incident import Incident

logger = logging.getLogger(__name__)

NON_ALPHANUMERIC_CHARACTERS: Final[re.Pattern[str]] = re.compile(r"[^\da-zA-Z]+")


@slack_client
def respond(
    body: dict[str, Any],
    text: str = "",
    blocks: str | Sequence[dict[str, Any] | Block] | None = None,
    client: WebClient = DefaultWebClient,
) -> None:
    """Respond to the user, depending on where the message was coming from."""
    user_id: str | None = body.get("user_id", body.get("user", {}).get("id"))
    channel_id: str | None = body.get("channel_id", body.get("channel", {}).get("id"))

    if not user_id and not channel_id:
        raise ValueError(
            "Cannot find user_id (or user.id) or channel_id (or channel.id) in the body. At least one is required."
        )

    # From Direct Message => Always respond on the conv between the user and the bot.
    if (body.get("channel_name") == "directmessage" or not channel_id) and user_id:
        # We should always be able to respond in this conversation...
        client.chat_postMessage(channel=user_id, text=text, blocks=blocks)
        return
    if not user_id:
        raise ValueError("Cannot find user_id (or user.id) in the body.")

    # From channel => post as ephemeral in the channel
    try:
        _send_ephemeral(text, blocks, client, user_id, channel_id)

    except (SlackApiError, ValueError):
        logger.warning(
            "Failed to send ephemeral chat message to user! Body: %s",
            body,
            exc_info=True,
        )
        # Fallback to DM
        if body.get("type") != "view_submission":
            client.chat_postMessage(
                channel=user_id,
                text=":warning: The bot could not respond in the channel you invoked it. Please add it to this channel or conversation if you want to interact with the bot there. If you believe this is a bug, please tell @pulse.",
            )
        client.chat_postMessage(channel=user_id, text=text, blocks=blocks)


def _send_ephemeral(
    text: str,
    blocks: str | Sequence[dict[str, Any] | Block] | None,
    client: WebClient,
    user_id: str,
    channel_id: str | None,
) -> None:
    if not channel_id:
        raise ValueError("Cannot find channel_id (or channel.id) in the body.")
    client.chat_postEphemeral(
        user=user_id, channel=channel_id, text=text, blocks=blocks
    )


def get_slack_user_id_from_body(body: dict[str, Any]) -> str | None:
    """Get the slack user id from the body of a Slack request, in `user_id` or `user.id`."""
    return body.get("user_id", body.get("user", {}).get("id"))


def channel_name_from_incident(incident: Incident) -> str:
    """Lowercase, truncated at 80 chars, this is obviously the channel #name."""
    if (
        not hasattr(incident, "created_at")
        or not hasattr(incident, "id")
        or not incident.created_at
        or not incident.id
    ):
        raise RuntimeError(
            "Incident must be saved before slack_channel_name can be computed"
        )
    date_formatted = localtime(incident.created_at).strftime("%Y%m%d")
    if incident.environment is not None and incident.environment.value != "PRD":
        topic = f"{date_formatted}-{str(incident.id)[:8]}-{incident.environment.value}-{incident.component.name}"
    else:
        topic = f"{date_formatted}-{str(incident.id)[:8]}-{incident.component.name}"

    # Strip non-alphanumeric characters, cut at 80 chars
    # XXX django.utils.text.slugify should be used instead
    topic = topic.replace(" ", "-")
    topic = NON_ALPHANUMERIC_CHARACTERS.sub("-", topic)
    return topic.lower()[:80]



from __future__ import annotations

import re
from typing import TypeAlias, TypedDict

from django.utils.timezone import (  # type: ignore[attr-defined]
    datetime,
    get_current_timezone,
)

from firefighter.incidents.views.date_utils import get_quarter_from_week

ConfluencePageId: TypeAlias = int | str
"""Alias of `int | str`"""


class ConfluenceContentVersionData(TypedDict, total=False):
    when: datetime | None
    friendlyWhen: str | None
    message: str | None
    number: int | None
    minorEdit: bool | None
    syncRev: str | None
    syncRevSource: str | None
    confRev: str | None
    contentTypeModified: bool | None


class ConfluenceContentLinksData(TypedDict, total=False):
    self: str
    tinyui: str | None
    editui: str | None
    webui: str | None


class ConfluencePage(TypedDict):
    id: ConfluencePageId
    type: str
    status: str | None
    title: str
    version: ConfluenceContentVersionData | None
    links: ConfluenceContentLinksData | None
    body: dict[str, str] | None


class PageInfo(TypedDict):
    """Dict mapped to mandatory fields for a [firefighter.confluence.models.ConfluencePage][]."""

    name: str
    page_id: str
    page_url: str
    page_edit_url: str


CONFLUENCE_TITLE = r"\#(?P<date>\d{8}-inc|\d{8})-(?P<id>\d{2,4})(-[a-zA-Z0-9-]+( |$)| )(?P<priority>\((?:(SEV|P)-?\s?[1-5]{1}\))|(?:\(GAMEDAY\)))?[ \t]?(?P<title>.*)?"
CONFLUENCE_PM_TITLE_REGEX = re.compile(CONFLUENCE_TITLE)

CONFLUENCE_PM_ARCHIVE_TITLE = r"^(?:\[Archive\] )?Q([1-4]) (\d{4})"
CONFLUENCE_PM_ARCHIVE_TITLE_REGEX = re.compile(CONFLUENCE_PM_ARCHIVE_TITLE)


def parse_postmortem_title(title: str) -> tuple[datetime, str] | tuple[None, None]:
    """Parse postmortem title and return the date and quarter associated to this date, in the format YYYYQX.

    Args:
        title (str): Title of the postmortem, in the format #YYYYMMDD-<id>-<component> (SEV/P<N>) <title>

    Returns:
        tuple[datetime, str] | tuple[None, None]: Date and quarter associated to this date.
    """
    match = CONFLUENCE_PM_TITLE_REGEX.match(title)
    if not match:
        return None, None

    res = match.groupdict()
    grp = res["date"]
    grp = grp.replace("-inc", "")

    date_ = datetime.strptime(grp, "%Y%m%d").replace(tzinfo=get_current_timezone())
    quarter = get_quarter_from_week(date_.isocalendar()[1])
    year = date_.isocalendar()[0]
    fmt = f"{year}Q{quarter}"
    return date_, fmt


def parse_postmortem_archive_title(title: str) -> re.Match[str] | None:
    return CONFLUENCE_PM_ARCHIVE_TITLE_REGEX.match(title)


def parse_runbook_title(title: str) -> str | None:
    """Parse runbooks title and return the service name, or None if the title is not valid.

    Useful to remove prefixes like [WIP] or [ARCHIVED] from the title, to sort runbooks.

    Args:
        title (str): Title of the runbook

    Returns:
        str | None: Service name or None if the title is not valid.
    """
    if "[RUNBOOK]" not in title or "[TEMPLATE]" in title.upper():
        return None

    # Remove leading text in brackets
    title = re.sub(r"^\[.*\] ", "", title)
    return title.strip()



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from slack_sdk.models.views import View

from firefighter.incidents.forms.update_roles import IncidentUpdateRolesForm
from firefighter.slack.slack_templating import slack_block_footer, slack_block_separator
from firefighter.slack.views.modals.base_modal.base import ModalForm
from firefighter.slack.views.modals.base_modal.mixins import (
    IncidentSelectableModalMixin,
)

if TYPE_CHECKING:
    from slack_bolt.context.ack.ack import Ack

    from firefighter.incidents.models.incident import Incident, User
    from firefighter.slack.views.modals.base_modal.form_utils import (
        SlackFormAttributesDict,
    )

logger = logging.getLogger(__name__)


class IncidentUpdateRolesFormSlack(IncidentUpdateRolesForm):
    slack_fields: SlackFormAttributesDict = {}


class UpdateRolesModal(
    IncidentSelectableModalMixin, ModalForm[IncidentUpdateRolesFormSlack]
):
    open_action: str = "open_modal_incident_update_roles"
    update_action: str = "update_modal_incident_update_roles"
    push_action: str = "push_modal_incident_update_roles"
    callback_id: str = "incident_update_roles"
    form_class = IncidentUpdateRolesFormSlack

    def build_modal_fn(self, incident: Incident, user: User, **kwargs: Any) -> View:
        blocks = self.get_form_class()(
            incident=incident,
            user=user,
        ).slack_blocks()
        blocks.append(slack_block_separator())
        blocks.append(slack_block_footer())
        return View(
            type="modal",
            title=f"Update roles for #{incident.id}"[:24],
            submit="Update roles"[:24],
            callback_id=self.callback_id,
            private_metadata=str(incident.id),
            blocks=blocks,
        )

    def handle_modal_fn(  # type: ignore
        self, ack: Ack, body: dict[str, Any], incident: Incident, user: User
    ):
        slack_form = self.handle_form_errors(
            ack,
            body,
            forms_kwargs={
                "incident": incident,
                "user": user,
            },
        )
        if slack_form is None:
            return
        form: IncidentUpdateRolesFormSlack = slack_form.form

        if len(form.cleaned_data) == 0:
            # XXX We should have a prompt for empty forms
            return
        form.save()


modal_update_roles = UpdateRolesModal()



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, ParamSpec, Self, TypedDict, TypeVar

from django.conf import settings
from slack_bolt.app.app import App
from slack_bolt.middleware.authorization.single_team_authorization import (
    SingleTeamAuthorization,
)

if TYPE_CHECKING:
    from collections.abc import Callable

    from slack_sdk.web.client import WebClient

logger = logging.getLogger(__name__)


class SlackAppDetails(TypedDict):
    url: str
    team_id: str
    team: str
    user: str
    user_id: str
    bot_id: str
    is_enterprise_install: bool


class SlackApp(App):
    """Subclass of the Slack App, as a singleton."""

    instance: App | None = None
    details: SlackAppDetails

    def __new__(cls, *args: Any, **kwargs: Any) -> Self:
        if not cls.instance:
            if settings.FF_SLACK_SKIP_CHECKS:
                logger.warning(
                    "Skipping Slack checks! Only use for testing or demo. Features related to the Bot User or the Workspace may not work (e.g. some generated URLs may be invalid)"
                )
                kwargs["token_verification_enabled"] = False
                kwargs["request_verification_enabled"] = False
                kwargs["ssl_check_enabled"] = False
                kwargs["url_verification_enabled"] = False

            slack_bot_token: str = settings.SLACK_BOT_TOKEN
            slack_signing_secret: str = settings.SLACK_SIGNING_SECRET
            kwargs["token"] = slack_bot_token
            kwargs["signing_secret"] = slack_signing_secret
            kwargs["ignoring_self_events_enabled"] = False
            cls.instance = App(*args, **kwargs)
            single_team_authorization = next(
                filter(
                    lambda m: isinstance(m, SingleTeamAuthorization),
                    cls.instance._middleware_list,  # noqa: SLF001
                )
            )

            if isinstance(single_team_authorization, SingleTeamAuthorization):
                res = single_team_authorization.auth_test_result
                if settings.FF_SLACK_SKIP_CHECKS:
                    cls.details = cls.instance.details = SlackAppDetails(  # type: ignore
                        url="",
                        team="",
                        user="",
                        team_id="",
                        user_id="",
                        bot_id="",
                        is_enterprise_install=False,
                    )
                    return cls.instance  # type: ignore[return-value]
                if res is None:
                    logger.critical("Could not verify Slack credentials! Exiting.")
                    raise RuntimeError("Could not verify Slack credentials! Exiting.")
                if not res.get("ok"):
                    logger.critical(
                        "Could not verify Slack credentials! Exiting. Credentials: %s",
                        res,
                    )
                    raise RuntimeError("Could not verify Slack credentials! Exiting.")
                cls.details = cls.instance.details = SlackAppDetails(  # type: ignore
                    url=res["url"],
                    team=res["team"],
                    user=res["user"],
                    team_id=res["team_id"],
                    user_id=res["user_id"],
                    bot_id=res["bot_id"],
                    is_enterprise_install=res["is_enterprise_install"],
                )
                logger.debug("SlackAppDetails: %s", cls.details)
        return cls.instance  # type: ignore[return-value]


P = ParamSpec("P")
R = TypeVar("R")


# Now we can type the decorator more accurately
def slack_client(function: Callable[P, R]) -> Callable[P, R]:
    """Adds a Slack client in `client` kwargs, if none is provided.

    Can be used as a decorator with `@slack_client` or as a function with `slack_client(function)`.
    """

    def wrap_function(*args: P.args, **kwargs: P.kwargs) -> R:
        if "client" not in kwargs:
            kwargs["client"] = SlackApp().client
        return function(*args, **kwargs)

    return wrap_function


DefaultWebClient: WebClient = None  # type: ignore



import logging

from django.db import migrations

logger = logging.getLogger(__name__)


def get_component_mappings() -> list:
    """
    Returns a list of tuples for updating existing component names and their attributes.

    Each tuple contains:
        - old_name (str): The current name of the component.
        - new_name (str): The new name to assign to the component.
        - slack_channel (str): The associated Slack channel for the component.
        - group_name (str): The name of the group to which the component belongs.

    Returns:
        list: A list of tuples, each representing the details for a component update.
    """
    return [
        ("Commercial Animation (Mabaya cat. integration & ad request, ...)", "Commercial Animation", "impact-commercial-animation", "Marketplace"),
        ("Mobile Apps", "Mobile Apps", "impact-mobile-apps", "Marketplace"),
        ("Spartacux Foundations", "Spartacux Foundations", "impact-spartacux-foundations", "Marketplace"),
        ("Tracking", "Tracking", "impact-tracking", "Marketplace"),
        ("HUB Integrators", "HUB Integrators", "impact-hub-integrators", "Seller"),
        ("Seller Account and Feeds", "Seller Catalog and Offer Management", "impact-seller-catalog-offer-management", "Seller"),
        ("Toolbox", "Seller Admin and Experience", "impact-seller-admin-experience", "Seller"),
        ("Seller Services (Mabaya BO, Subscriptions, MF)", "Seller Services", "impact-seller-services", "Seller"),
        ("Catalog Performance", "Catalog Performance", "impact-catalog-performance", "Catalog"),
        ("Offer (Price, Stock)", "Offer (Price & Stock)", "impact-offer-price-stock", "Catalog"),
        ("Back Office", "BO Catalog - Master Experience", "impact-bo-catalog-master-experience", "Catalog"),
        ("Order Lifecycle", "Order management", "impact-order-management", "Operations"),
        ("Delivery Experience", "Delivery experience", "impact-delivery-experience", "Operations"),
        ("Cloud Infrastructure", "Cloud Infrastructure", "impact-cloud-infrastructure", "Platform"),
        ("Spinak", "Spinak", "impact-spinak", "Platform"),
        ("CDN", "CDN", "impact-cdn", "Platform"),
        ("Gitlab", "Gitlab", "impact-gitlab", "Platform"),
    ]


def get_new_components() -> dict:
    """
    Returns a dictionary of new components to be created.

    Each entry in the dictionary maps a component name to a tuple containing:
    - group_name: The name of the group the component belongs to.
    - slack_channel: The associated Slack channel for the component.

    Returns:
        dict: A mapping of component names to (group name, slack channel) tuples.
    """
    return {
        "Traffic acquisition": ("Marketing & Communication", "impact-traffic-acquisition"),
        "Company reputation": ("Marketing & Communication", "impact-company-reputation"),
        "Loyalty and coupons": ("Payment Operations", "impact-loyalty-coupons"),
        "Payouts to seller": ("Payment Operations", "impact-payouts-to-seller"),
        "Refunds": ("Payment Operations", "impact-refunds"),
        "Returns": ("Operations", "impact-returns"),
        "Customer service": ("Operations", "impact-customer-service"),
        "Inventory": ("Operations", "impact-inventory"),
        "VAT": ("Finance", "impact-vat"),
        "Seller's invoices": ("Finance", "impact-sellers-invoices"),
        "Customer's invoices": ("Finance", "impact-customers-invoices"),
        "Accounting": ("Finance", "impact-accounting"),
        "Revenue": ("Finance", "impact-revenue"),
        "Compromised laptop / server": ("Security", "impact-compromised-laptop-server"),
    }


def add_new_components(apps, schema_editor):
    Component = apps.get_model("incidents", "Component")
    Group = apps.get_model("incidents", "Group")
    new_components = get_new_components()

    for name, (group_name, _slack) in new_components.items():
        logger.info(f"Creating new component: '{name}' belonging to group '{group_name}'")
        try:
            group_instance = Group.objects.get(name=group_name)
            new_component = Component(name=name, group=group_instance)
            new_component.save()
        except Exception:
            logger.exception(f"Failed to create new group: '{group_name}'.")


def remove_new_components(apps, schema_editor):
    Component = apps.get_model("incidents", "Component")
    new_component_names = get_new_components().keys()

    for name in new_component_names:
        try:
            component = Component.objects.get(name=name)
            logger.info(f"Removing component: '{name}'")
            component.delete()
        except Exception:
            logger.exception(f"Component '{name}' does not exist, skipping removal.")


def update_component_names(apps, schema_editor):
    Component = apps.get_model("incidents", "Component")
    Group = apps.get_model("incidents", "Group")
    component_mappings = get_component_mappings()

    updated_count = 0

    for old_name, new_name, _slack_channel, group_name in component_mappings:
        try:
            component = Component.objects.get(name=old_name)
            logger.info(f"Updating: '{old_name}' to '{new_name}'")
            component.name = new_name

            # WARN: this operaration is impossible to revert
            group_instance = Group.objects.get(name=group_name)
            component.group = group_instance
            component.save()
            updated_count += 1
        except Exception:
            logger.exception(f"Component '{old_name}' does not exist, cannot proceed with updates.")


def revert_component_names(apps, schema_editor):
    Component = apps.get_model("incidents", "Component")
    reverse_mappings = {new_name: old_name for old_name, new_name, _, _ in get_component_mappings()}

    updated_count = 0

    for new_name, old_name in reverse_mappings.items():

        try:
            component = Component.objects.get(name=new_name)
            logger.info(f"Restoring '{new_name}' back to '{old_name}'")
            component.name = old_name
            component.save()
            updated_count += 1
        except Exception:
            logger.exception(f"Component '{new_name}' does not exist, skipping restoration.")


class Migration(migrations.Migration):

    dependencies = [
        ("incidents", "0006_update_group_names"),
    ]

    operations = [
        migrations.RunPython(update_component_names, revert_component_names),
        migrations.RunPython(add_new_components, remove_new_components),
    ]



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from hypothesis import given
from hypothesis.extra import django
from hypothesis.strategies import builds

from firefighter.incidents.factories import IncidentFactory

if TYPE_CHECKING:
    from firefighter.incidents.models import Incident


@pytest.mark.django_db
class TestIncident(django.TestCase):
    """This is a property-based test that ensures model correctness."""

    @given(builds(IncidentFactory.build))
    def test_model_properties(self, instance: Incident) -> None:
        """Tests that instance can be saved and has correct representation."""
        instance.component.group.save()
        instance.component.save()
        instance.environment.save()
        instance.priority.save()
        instance.created_by.save()
        instance.save()

        assert instance.id > 0



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import django_tables2 as tables
from django.conf import settings

from firefighter.firefighter.tables_utils import BASE_TABLE_ATTRS
from firefighter.incidents.models import Component, Incident
from firefighter.incidents.models.incident import IncidentStatus

if TYPE_CHECKING:
    from datetime import timedelta


class IncidentTable(tables.Table):
    class Meta:
        model = Incident
        template_name = "incidents/table.html"
        fields = (
            "id",
            "title",
            "priority",
            "status",
            "environment",
            "component",
            "component__group",
            "created_at",
        )
        order_by = "-id"
        attrs: dict[str, str | dict[str, str]] = {
            **BASE_TABLE_ATTRS,
            "aria-label": "Incident list table",
        }

    id = tables.Column(
        linkify=True,
        attrs={"a": {"class": "text-primary hover:text-primary/80"}},
    )
    title = tables.Column(
        linkify=True,
        attrs={
            "a": {"class": "overflow-ellipsis"},
            "td": {"class": "px-3 py-4 max-w-sm md:max-w-md lg:max-w-lg truncate"},
            "th": {
                "class": "px-3 py-3 tracking-wider hover:underline uppercase text-xs font-medium text-left"
            },
        },
    )

    priority = tables.TemplateColumn(
        template_name="incidents/table/priority_column.html"
    )
    status = tables.TemplateColumn(
        template_name="incidents/table/status_column.html",
        attrs={"td": {"class": "px-3 py-4 whitespace-nowrap flex justify-around"}},
        extra_context={"IncidentStatus": IncidentStatus},
        order_by=("_status",),
    )
    created_at = tables.DateTimeColumn(
        short=True,
        format=settings.SHORT_DATETIME_FORMAT,
        attrs={"td": {"class": "table-td text-center"}},
    )
    environment = tables.Column(attrs={"td": {"class": "table-td text-center"}})
    component = tables.Column(attrs={"td": {"class": "table-td text-center"}})
    component__group = tables.Column(attrs={"td": {"class": "table-td text-center"}})


class ComponentsTable(tables.Table):
    class Meta:
        model = Component
        template_name = "incidents/table.html"
        fields = (
            "name",
            "group__name",
            "mtbf",
            "incident_count",
        )
        attrs: dict[str, str | dict[str, str]] = {
            **BASE_TABLE_ATTRS,
            "aria-label": "Incident list table",
        }

    name = tables.Column(
        linkify=True,
        attrs={
            "td": {"class": "table-td text-left"},
            "a": {
                "class": "text-primary hover:text-indigo-900 dark:text-indigo-400 dark:hover:text-indigo-200"
            },
        },
    )

    @staticmethod
    def render_mtbf(record: Component, *args: Any) -> str:
        mtbf: timedelta | None = record.mtbf  # type: ignore[attr-defined]
        return str(mtbf).split(".", maxsplit=1)[0] if mtbf else "N/A"

    @staticmethod
    def render_incident_count(record: Component, *args: Any) -> int:
        return int(record.incident_count) if record.incident_count else 0  # type: ignore[attr-defined]

    group__name = tables.Column(verbose_name="Group")



# Generated by Django 4.2.7 on 2023-12-04 11:07

import uuid

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    replaces = [
        ("confluence", "0001_squashed_0007"),
        ("confluence", "0008_alter_runbook_service_name_and_more"),
        ("confluence", "0009_confluencepage_version_and_more"),
    ]

    dependencies = [
        ("incidents", "0030_status_enum_constraints"),
        ("incidents", "0006_alter_user_slack_id"),
        ("incidents", "0005_alter_incidentupdate_communication_lead"),
    ]

    operations = [
        migrations.CreateModel(
            name="ConfluencePage",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("name", models.CharField(max_length=255)),
                ("page_id", models.PositiveBigIntegerField(db_index=True, unique=True)),
                ("page_url", models.URLField(max_length=1024)),
                (
                    "created_at",
                    models.DateTimeField(
                        auto_now_add=True,
                        help_text="When was the DB record created, not when the page was created on Confluence.",
                    ),
                ),
                (
                    "updated_at",
                    models.DateTimeField(
                        auto_now=True,
                        help_text="When was the DB record updated, not when the page was created on Confluence.",
                    ),
                ),
                ("page_edit_url", models.URLField(max_length=1024)),
                ("body_export_view", models.TextField(blank=True, null=True)),
                ("body_storage", models.TextField(blank=True, null=True)),
                ("body_view", models.TextField(blank=True, null=True)),
                ("version", models.JSONField(default=dict)),
            ],
        ),
        migrations.CreateModel(
            name="PostMortem",
            fields=[
                (
                    "confluencepage_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="confluence.confluencepage",
                    ),
                ),
                (
                    "incident",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="postmortem_for",
                        to="incidents.incident",
                    ),
                ),
            ],
            bases=("confluence.confluencepage",),
        ),
        migrations.CreateModel(
            name="Runbook",
            fields=[
                (
                    "confluencepage_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="confluence.confluencepage",
                    ),
                ),
                ("service_name", models.CharField(max_length=255)),
                ("service_type", models.CharField(max_length=255)),
                ("title", models.CharField(max_length=255)),
            ],
            bases=("confluence.confluencepage",),
        ),
    ]



"""PagerDuty Celery tasks."""

from __future__ import annotations

from firefighter.pagerduty.tasks.fetch_oncall import fetch_oncalls
from firefighter.pagerduty.tasks.fetch_services import fetch_services
from firefighter.pagerduty.tasks.fetch_users import fetch_users



from __future__ import annotations

import logging

from celery import shared_task

from firefighter.confluence.service import confluence_service
from firefighter.confluence.utils import (
    ConfluencePage,
    ConfluencePageId,
    parse_runbook_title,
)

logger = logging.getLogger(__name__)


@shared_task(name="confluence.sort_runbooks", soft_time_limit=60 * 5, time_limit=60 * 7)
def sort_runbooks(*, dry_run: bool = False) -> None:
    """Sort runbooks in Confluence.

    Args:
        dry_run (bool, optional): Should actions be performed? If True, Confluence will be accessed in read-only mode. Defaults to False.
    """
    logger.info(f"Running in {'dry run' if dry_run else 'real'} mode")

    # 1. Get top level pages
    folders = _get_top_level_pages(confluence_service.RUNBOOKS_FOLDER_ID)

    # 2. Sort runbooks in each folder
    sort_runbooks_in_folders(folders, dry_run=dry_run)


def _get_top_level_pages(
    root_page_id: ConfluencePageId,
) -> list[ConfluencePageId]:
    """Get all top level pages in the runbooks folder."""
    folders: list[ConfluencePageId] = []
    children_pages = confluence_service.get_page_children_pages(root_page_id)

    for page in children_pages:
        title: str = page["title"]
        if "TEMPLATE" in title.upper():
            continue
        folders.append(int(page["id"]))
        logger.debug(f"Found runbook folder {title} with id {page['id']}")

    logger.info(f"Found {len(folders)} runbooks folder ({folders})")
    return folders


def sort_runbooks_in_folders(
    folders: list[ConfluencePageId], *, dry_run: bool = False
) -> None:
    """Sort runbooks in each folder.

    Args:
        folders (list[ConfluencePageId]): List of folders to sort runbooks in.
        dry_run (bool, optional): Do not perform the sort. Defaults to False.
    """
    for folder_page_id in folders:
        # Get children per archive page
        children: list[ConfluencePage] = confluence_service.get_page_children_pages(
            folder_page_id, expand=""
        )

        folder_children: list[tuple[ConfluencePageId, str, ConfluencePage]] = []

        for page in children:
            title: str = page["title"]
            page_id: ConfluencePageId = page["id"]

            runbook_name_clean = parse_runbook_title(title)
            if runbook_name_clean:
                folder_children.append((page_id, runbook_name_clean, page))

        sorted_children = sorted(folder_children, key=lambda x: x[1].lower())
        # Check if already sorted
        if sorted_children == folder_children:
            logger.debug(f"Folder {folder_page_id} is already sorted. Skipping")
            continue
        clean_sorted_children: list[tuple[ConfluencePageId, ConfluencePage]] = [
            (x[0], x[2]) for x in sorted_children
        ]
        logger.info(f"Folder {folder_page_id} is not sorted. Sorting")
        confluence_service.sort_pages(clean_sorted_children, dry_run=dry_run)



from __future__ import annotations

import json
import logging
import re
from typing import TYPE_CHECKING, Any, Never, cast

from django import forms
from django.conf import settings
from django.db.models import Model, QuerySet
from slack_sdk.models.blocks.basic_components import MarkdownTextObject
from slack_sdk.models.blocks.blocks import (
    ContextBlock,
    DividerBlock,
    SectionBlock,
)
from slack_sdk.models.views import View

from firefighter.firefighter.utils import get_in
from firefighter.incidents.forms.select_impact import SelectImpactForm
from firefighter.incidents.models.impact import ImpactType
from firefighter.incidents.models.priority import Priority
from firefighter.slack.slack_app import SlackApp
from firefighter.slack.views.modals import modal_open
from firefighter.slack.views.modals.base_modal.base import ModalForm
from firefighter.slack.views.modals.base_modal.form_utils import (
    SlackForm,
    SlackFormAttributesDict,
    SlackFormJSONEncoder,
)
from firefighter.slack.views.modals.base_modal.mixins import (
    IncidentSelectableModalMixin,
)
from firefighter.slack.views.modals.base_modal.modal_utils import (
    push_modal,
    update_modal,
)
from firefighter.slack.views.modals.opening.types import OpeningData, ResponseType

if TYPE_CHECKING:
    from slack_bolt import BoltRequest
    from slack_bolt.context.ack.ack import Ack

    from firefighter.incidents.models.impact import ImpactLevel
    from firefighter.incidents.models.user import User

app = SlackApp()
logger = logging.getLogger(__name__)
BASE_URL: str = settings.BASE_URL
SLACK_APP_EMOJI: str = settings.SLACK_APP_EMOJI


class SelectImpactFormSlack(SelectImpactForm):
    slack_fields: SlackFormAttributesDict = {"customer_impact": {}}


class SelectImpactModal(
    IncidentSelectableModalMixin,
    ModalForm[SelectImpactFormSlack],
):
    """TODO: The detailed impacts selected should be saved on the incident."""

    open_action: str = "open_incident_select_impact"
    open_shortcut: str = "open_incident_select_impact"
    push_action: str = "push_incident_select_impact"
    callback_id: str = "incident_select_impact"

    form_class: type[SelectImpactFormSlack] = SelectImpactFormSlack

    def build_modal_fn(
        self, body: dict[str, Any], open_incident_context: OpeningData, **kwargs: Never
    ) -> View:
        initial_form_values_impact = open_incident_context.get("impact_form_data", {})
        if initial_form_values_impact:
            for field_name, field in self.form_class().fields.items():
                if (
                    isinstance(field, forms.ModelChoiceField)
                    and field_name in initial_form_values_impact
                    and not isinstance(initial_form_values_impact[field_name], Model)
                    and field.queryset is not None
                    and hasattr(field.queryset.model, "DoesNotExist")
                ):
                    try:
                        initial_form_values_impact[field_name] = field.queryset.get(
                            pk=initial_form_values_impact[field_name]
                        )
                    except field.queryset.model.DoesNotExist:
                        initial_form_values_impact[field_name] = None  # type: ignore
        blocks = [
            SectionBlock(
                text=f"{SLACK_APP_EMOJI} Define the impact(s) to find the incident priority (<https://manomano.atlassian.net/wiki/spaces/TC/pages/4024500413/Priority+levels|documentation>):"
            ),
            *self.get_form_class()(initial=initial_form_values_impact).slack_blocks(
                "section_accessory"
            ),
        ]
        priority_data = open_incident_context.get("priority")

        if open_incident_context and priority_data:
            if not isinstance(priority_data, Priority):
                priority = Priority.objects.get(pk=priority_data)
            elif isinstance(priority_data, Priority):
                priority = priority_data
            else:
                err_msg = f"Invalid priority data: {priority_data}"  # type: ignore[unreachable]
                raise TypeError(err_msg)
            process = ":slack: Slack :jira_new: Jira ticket" if open_incident_context.get("response_type") == "critical" else ":jira_new: Jira ticket"

            impact_descriptions = self.extract_descriptions(open_incident_context)

            blocks.extend((
                DividerBlock(),
                ContextBlock(
                    elements=[
                        MarkdownTextObject(
                            text=(
                                f":dart: Selected priority: *{priority} - {priority.description}*\n"
                                f"⏱️ SLA: {priority.sla}\n"
                                f":gear: Process: {process}\n"
                                f":pushpin: Selected impacts:\n"
                                f"{impact_descriptions}\n"
                            )
                        )
                    ]
                )
            ))

        return View(
            type="modal",
            callback_id=self.callback_id,
            title="Incident Impact"[:24],
            private_metadata=str(body["trigger_id"]),
            blocks=blocks,
            submit="Save impacts",
        )

    def extract_descriptions(self, open_incident_context: OpeningData) -> str:
        impact_form_data = open_incident_context.get("impact_form_data", {})
        impact_descriptions = ""
        if impact_form_data:
            for value in impact_form_data.values():
                if value.name != "NO" and value.description:
                    if hasattr(value, "impact_type_id") and value.impact_type_id:
                        impact_type = ImpactType.objects.get(pk=value.impact_type_id)
                        if impact_type:
                            impact_descriptions += f"\u00A0\u00A0 :exclamation: {impact_type} - {value}\n"
                    for line in str(value.description).splitlines():
                        impact_descriptions += f"\u00A0\u00A0\u00A0\u00A0\u00A0\u00A0 • {line}\n"
        return impact_descriptions

    def handle_modal_fn(  # type: ignore
        self,
        ack: Ack,
        body: dict[str, Any],
        user: User | None = None,
    ) -> None:
        form = self.handle_form_errors(ack, body, {})
        if form is None:
            return

        private_metadata = self._update_private_metadata(body, form)
        view = modal_open.build_modal_fn(
            open_incident_context=private_metadata, user=user
        )
        view.private_metadata = json.dumps(private_metadata, cls=SlackFormJSONEncoder)
        update_modal(
            view=view,
            trigger_id=get_in(body, ["trigger_id"]),
            view_id=get_in(body, ["view", "root_view_id"]),
        )

    def _handle_action_push(
        self,
        request: BoltRequest,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        body = request.body
        data = cast(
            "OpeningData", json.loads(body.get("actions", [{}])[0].get("value", {})) or {}
        )
        view = self.build_modal_fn(body, open_incident_context=data)
        request.context.ack()

        view.private_metadata = json.dumps(data, cls=SlackFormJSONEncoder)
        push_modal(view=view, body=body)

    @app.block_action(re.compile(r"set_impact_type.+"))
    @staticmethod
    def handle_select_impact(ack: Ack, body: dict[str, Any], **kwargs: Any) -> None:
        modal = SelectImpactModal()
        form = modal.handle_form_errors(ack, body, {})
        if form is None:
            return

        private_metadata = SelectImpactModal._update_private_metadata(body, form)
        view = modal.build_modal_fn(body, open_incident_context=private_metadata)
        view.private_metadata = json.dumps(private_metadata, cls=SlackFormJSONEncoder)
        update_modal(
            view=view,
            body=body,
        )

    @staticmethod
    def _calculate_proposed_incident_type(
        suggested_priority_value: int,
    ) -> ResponseType:
        return "critical" if suggested_priority_value <= 3 else "normal"

    @staticmethod
    def _update_private_metadata(
        body: dict[str, Any], form: SlackForm[SelectImpactFormSlack]
    ) -> OpeningData:
        private_metadata_raw: dict[str, Any] = json.loads(
            body.get("view", {}).get("private_metadata", {})
        )

        # Convert the form fields str to Models
        for field_name, field in form.form.fields.items():
            if (
                isinstance(field, forms.ModelChoiceField)
                and field_name in form.form.data
                and not isinstance(form.form.data[field_name], Model)
                and field.queryset is not None
            ):
                queryset = cast("QuerySet[ImpactLevel]", field.queryset)
                try:
                    form.form.data[field_name] = queryset.get(  # type: ignore
                        pk=form.form.data[field_name]
                    )
                except queryset.model.DoesNotExist:
                    form.form.data[field_name] = None  # type: ignore
        return OpeningData(
            priority=Priority.objects.get(
                value=form.form.suggest_priority_from_impact()
            ),
            response_type=SelectImpactModal._calculate_proposed_incident_type(
                form.form.suggest_priority_from_impact()
            ),
            impact_form_data=cast("dict[str, Any]", form.form.data),
            details_form_data=private_metadata_raw.get("details_form_data", {}),
            incident_type=private_metadata_raw.get("incident_type"),
        )


modal_select_impact = SelectImpactModal()



from __future__ import annotations

import logging
import re
from typing import TYPE_CHECKING, Any, cast

from django.conf import settings
from django.contrib.auth.mixins import LoginRequiredMixin
from django.db.models.expressions import OuterRef, Subquery
from django.db.models.query import Prefetch
from django.http import HttpRequest, HttpResponse, HttpResponseRedirect
from django.urls import reverse, reverse_lazy
from django.views import generic
from django_filters.views import FilterView
from django_tables2.views import SingleTableMixin

from firefighter.firefighter.views import CustomDetailView
from firefighter.incidents.enums import IncidentStatus
from firefighter.incidents.forms import CreateIncidentForm
from firefighter.incidents.forms.update_key_events import IncidentUpdateKeyEventsForm
from firefighter.incidents.models.impact import Impact
from firefighter.incidents.models.incident import Incident, IncidentFilterSet
from firefighter.incidents.models.incident_update import IncidentUpdate
from firefighter.incidents.models.metric_type import IncidentMetric
from firefighter.incidents.signals import (
    create_incident_conversation,
    incident_key_events_updated,
)
from firefighter.incidents.tables import IncidentTable
from firefighter.incidents.views.reports import weekly_dashboard_context

if TYPE_CHECKING:
    from firefighter.firefighter.utils import HtmxHttpRequest

logger = logging.getLogger(__name__)


class IncidentListView(SingleTableMixin, FilterView):
    table_class = IncidentTable
    context_object_name = "incidents"
    filterset_class = IncidentFilterSet
    model = Incident

    paginate_by = 125
    paginate_orphans = 15
    queryset = Incident.objects.select_related(
        "priority", "component__group", "environment"
    ).order_by("-id")

    def get_template_names(self) -> list[str]:
        request = cast("HtmxHttpRequest", self.request)
        if request.htmx and not request.htmx.boosted:
            template_name = "layouts/partials/partial_table_list_paginated.html"
        else:
            template_name = "pages/incident_list.html"

        return [template_name]

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        """No *args to pass."""
        # Call the base implementation first to get a context
        context = super().get_context_data(**kwargs)
        context["page_range"] = context["table"].paginator.get_elided_page_range(
            context["table"].page.number, on_each_side=2, on_ends=1
        )
        context["filter_order"] = [
            "search",
            "created_at",
            "status",
            "environment",
            "priority",
            "component",
        ]

        context["page_title"] = "Incidents List"
        context["api_url_export"] = reverse("api:incidents-list")
        return context


class IncidentStatisticsView(FilterView):
    context_object_name = "incidents_filtered"
    filterset_class = IncidentFilterSet
    model = Incident
    queryset = (
        Incident.objects.select_related("priority", "component__group", "environment")
        .all()
        .order_by("-id")
    )

    def get_template_names(self) -> list[str]:
        request = cast("HtmxHttpRequest", self.request)
        if request.htmx and not request.htmx.boosted:
            template_name = "pages/incident_statistics_partial.html"
        else:
            template_name = "pages/incident_statistics.html"
        return [template_name]

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        """No *args to pass."""
        context = super().get_context_data(**kwargs)
        context["page_title"] = "Statistics"
        context["filter_order"] = [
            "search",
            "created_at",
            "status",
            "environment",
            "priority",
            "component",
        ]
        context_data = weekly_dashboard_context(
            self.request, context.get("incidents_filtered", [])
        )
        return {**context, **context_data}


class DashboardView(generic.ListView[Incident]):
    template_name = "pages/dashboard.html"
    context_object_name = "incidents"
    sub = (
        IncidentUpdate.objects.filter(incident_id=OuterRef("id"))
        .order_by("-event_ts")
        .values("event_ts")[:1]
    )
    queryset = (
        Incident.objects.filter(_status__lt=IncidentStatus.CLOSED)
        .select_related("priority", "component__group", "environment", "created_by")
        .order_by("_status", "priority__value")
        .annotate(latest_event_ts=Subquery(sub.values("event_ts")))
    )

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        """No *args to pass."""
        # Call the base implementation first to get a context
        context = super().get_context_data(**kwargs)
        context["page_title"] = "Incidents Dashboard"
        return context


class IncidentDetailView(CustomDetailView[Incident]):
    template_name = "pages/incident_detail.html"
    context_object_name: str = "incident"
    pk_url_kwarg = "incident_id"
    model = Incident
    select_related = [
        "priority",
        "environment",
        "component__group",
        "conversation",
        "created_by",
    ]
    if settings.ENABLE_CONFLUENCE:
        select_related.append("postmortem_for")
    queryset = Incident.objects.select_related(*select_related).prefetch_related(
        Prefetch(
            "incidentupdate_set",
            queryset=IncidentUpdate.objects.select_related(
                "priority",
                "component__group",
                "created_by__slack_user",
                "created_by",
                "commander",
                "communication_lead",
            ).order_by("-event_ts"),
        ),
        Prefetch(
            "metrics", queryset=IncidentMetric.objects.select_related("metric_type")
        ),
        Prefetch(
            "impacts",
            queryset=Impact.objects.select_related("impact_type", "impact_level"),
        ),
    )

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        context = super().get_context_data(**kwargs)

        incident: Incident = context["incident"]

        additional_context = {
            "page_title": f"Incident #{incident.id}",
        }

        return {**context, **additional_context}


class IncidentCreateView(LoginRequiredMixin, generic.edit.FormView[CreateIncidentForm]):
    login_url = "/oidc/authenticate?fail=/admin/login/"
    form_class = CreateIncidentForm
    template_name = "pages/incident_create.html"

    def form_valid(self, form: CreateIncidentForm) -> ProcessAfterResponse:
        incident = Incident.objects.declare(
            **form.cleaned_data,
            created_by=self.request.user,
        )
        # Redirect to the new incident page:
        return ProcessAfterResponse(
            reverse("incidents:incident-detail", args=(incident.id,)),
            data={
                "incident": incident,
                "user": self.request.user,
            },
        )

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        context = super().get_context_data(**kwargs)
        context["page_title"] = "Declare Critical Incident"
        return context


class IncidentUpdateKeyEventsView(
    generic.detail.SingleObjectMixin[Incident],
    LoginRequiredMixin,
    generic.FormView[IncidentUpdateKeyEventsForm],
):
    form_class = IncidentUpdateKeyEventsForm
    context_object_name = "incident"
    pk_url_kwarg = "incident_id"
    success_url = reverse_lazy("incidents:incident-list")
    model = Incident
    object: Incident

    def get_template_names(self) -> list[str]:
        request = cast("HtmxHttpRequest", self.request)
        if request.htmx and not request.htmx.boosted:
            template_name = (
                "layouts/partials/incident_update_key_events_view_modal.html"
            )

        else:
            template_name = "pages/incident_update_key_events_form.html"
        return [template_name]

    def get_form(
        self, form_class: type[IncidentUpdateKeyEventsForm] | None = None
    ) -> IncidentUpdateKeyEventsForm:
        """Replace Markdown bold syntax with HTML bold syntax in form labels."""
        form = super().get_form(form_class)
        for field in form.fields:
            # Replace *ABC* with <bold>ABC</bold>
            form.fields[field].label = re.sub(
                r"\*(.*?)\*", r"<b>\1</b>", str(form.fields[field].label)
            )
        return form

    def get(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        self.object = self.get_object()
        return super().get(request, *args, **kwargs)

    def post(self, request: HttpRequest, *args: str, **kwargs: Any) -> HttpResponse:
        self.object = self.get_object()
        return super().post(request, *args, **kwargs)

    def form_valid(self, form: IncidentUpdateKeyEventsForm) -> HttpResponse:
        # This method is called when valid form data has been POSTed.
        # It should return an HttpResponse.

        form.save()
        # Send signal so we can update the Slack message if applicable
        logger.debug("Sending signal incident_key_events_updated")
        incident_key_events_updated.send_robust(
            __name__,
            incident=form.incident,
        )
        form.incident.compute_metrics()
        return super().form_valid(form)

    def get_success_url(self) -> str:
        return reverse("incidents:incident-detail", args=(self.object.id,))

    def get_form_kwargs(self) -> dict[str, Any]:
        return super().get_form_kwargs() | {
            "incident": self.object,
            "user": self.request.user,
        }


class ProcessAfterResponse(HttpResponseRedirect):
    """Custom Response, to trigger the Slack workflow after creating the incident and returning HTTP 201.

    TODO This does not work, the workflow is triggered before the response is sent. We need to do a celery task!
    TODO We need to redirect to the incident page or Slack conversation.
    """

    def __init__(
        self, redirect_to: str, data: dict[str, Any], *args: Any, **kwargs: Any
    ) -> None:
        super().__init__(redirect_to, *args, **kwargs)
        self.data = data

    def close(self) -> None:
        super().close()
        incident = self.data["incident"]
        create_incident_conversation.send(
            "create-incident",
            incident=incident,
        )



from __future__ import annotations

import logging
import uuid
from typing import TYPE_CHECKING, Any

from django.db import models
from django_stubs_ext.db.models import TypedModelMeta

from firefighter.firefighter.utils import get_first_in, get_in
from firefighter.incidents.models import Component, User
from firefighter.slack.slack_app import DefaultWebClient, SlackApp, slack_client

if TYPE_CHECKING:
    from collections.abc import Sequence  # noqa: F401

    from django_stubs_ext.db.models.manager import RelatedManager  # noqa: F401
    from slack_sdk.web.client import WebClient
logger = logging.getLogger(__name__)


class UserGroupManager(models.Manager["UserGroup"]):
    @staticmethod
    @slack_client
    def fetch_usergroup(
        group_slack_id: str | None = None,
        group_handle: str | None = None,
        client: WebClient = DefaultWebClient,
        **kwargs: Any,
    ) -> UserGroup | None:
        """Import a "usergroup" from Slack and return its UserGroup model.
        Either group_slack_id or group_handle must be provided.

        Args:
            group_slack_id (str | None, optional): The Slack usergroup id of the usergroup to import. Usually starts with `S`. Defaults to None.
            group_handle (str | None, optional): Handle (@xxx) of the usergroup to import. Defaults to None.
            client (WebClient, optional): Slack client. Defaults to DefaultWebClient.
            **kwargs: Additional keyword arguments to pass to created UserGroup model.

        Returns:
            UserGroup | None: UserGroup model or None if not found
        """
        slack_response_usergroups = UserGroupManager.fetch_all_usergroups_data(
            client=client
        )

        usergroup = UserGroupManager.get_usergroup_data_from_list(
            usergroups=slack_response_usergroups,
            group_slack_id=group_slack_id,
            group_handle=group_handle,
        )
        if not usergroup:
            logger.warning(
                f"Could not find matching group! group_slack_id: {group_slack_id}; group_handle: {group_handle}"
            )
            return None
        logger.debug(usergroup)

        return UserGroup(
            **UserGroupManager.parse_slack_response(usergroup),
            **kwargs,
        )

    @staticmethod
    def parse_slack_response(slack_usergroup: dict[str, Any]) -> dict[str, Any]:
        name = get_in(slack_usergroup, "name")
        handle = get_in(slack_usergroup, "handle")
        usergroup_id = get_in(slack_usergroup, "id")
        description = get_in(slack_usergroup, "description")
        is_external = get_in(slack_usergroup, "is_external")

        return {
            "name": name,
            "handle": handle,
            "usergroup_id": usergroup_id,
            "description": description,
            "is_external": is_external,
        }

    @staticmethod
    def fetch_all_usergroups_data(
        client: WebClient = DefaultWebClient,
        *,
        include_users: bool = False,
    ) -> list[dict[str, Any]]:
        """Fetch all usergroups from firefighter.slack.

        Returns the list of usergroups
        """
        slack_response_usergroups = client.usergroups_list(include_users=include_users)

        ug_list = get_in(slack_response_usergroups, "usergroups")
        if not isinstance(ug_list, list):
            err_msg = (
                f"Expected usergroups to be a list, but got {type(ug_list)}: {ug_list}"
            )

            raise TypeError(err_msg)
        return ug_list

    @staticmethod
    def get_usergroup_data_from_list(
        usergroups: list[dict[str, Any]],
        group_slack_id: str | None = None,
        group_handle: str | None = None,
    ) -> dict[str, Any] | None:
        if group_slack_id is None and group_handle is None:
            raise ValueError("Either group_slack_id or group_handle must be provided.")

        if group_slack_id:
            usergroup = get_first_in(get_in(usergroups, []), "id", (group_slack_id,))
        elif group_handle:
            usergroup = get_first_in(
                get_in(usergroups, []),
                "handle",
                (group_handle,),
            )
        else:
            return None

        return usergroup

    @staticmethod
    def fetch_usergroup_data(
        group_slack_id: str | None = None, group_handle: str | None = None
    ) -> dict[str, Any] | None:
        usergroups_data = UserGroupManager.fetch_all_usergroups_data()
        return UserGroupManager.get_usergroup_data_from_list(
            usergroups_data, group_slack_id, group_handle
        )


class UserGroup(models.Model):
    """Model a Slack API UserGroup.
    Reference: https://api.slack.com/types/usergroup.
    """

    objects: UserGroupManager = UserGroupManager()

    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)

    name = models.CharField(
        max_length=80,
        blank=True,
        help_text="Friendly name of the group. Corresponds to the `name` field in the Slack API. Max 80 characters.",
    )
    handle = models.CharField(
        max_length=80,
        blank=True,
        help_text="Indicates the value used to notify group members via a mention. The handle does not include the leading @. Corresponds to the `handle` field in the Slack API. Max 80 characters.",
    )
    description = models.CharField(
        max_length=140,
        blank=True,
        help_text="A short description of the group. Corresponds to the `description` field in the Slack API. Max 140 characters.",
    )
    usergroup_id = models.CharField(
        max_length=32,
        blank=True,
        null=True,
        unique=True,
        help_text="The Slack usergroup ID of the usergroup. Usually starts with `S`.",
    )

    is_external = models.BooleanField(
        default=False,
        help_text="Is this an external group, from an external Slack Workspace? Corresponds to the `is_external` field in the Slack API.",
    )

    components = models.ManyToManyField["Component", "Component"](
        Component,
        related_name="usergroups",
        blank=True,
        help_text="Incident created with this usergroup automatically add the group members to these issue categories.",
    )

    tag = models.CharField(
        max_length=80,
        blank=True,
        help_text="Used by FireFighter internally to mark special user groups (e.g. @team-secu, @team-incidents...). Must be empty or unique.",
    )

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    members = models.ManyToManyField(User, blank=True)

    class Meta(TypedModelMeta):
        verbose_name = "Slack user group"
        verbose_name_plural = "Slack user groups"

    def __str__(self) -> str:
        return f"@{self.handle} ({self.usergroup_id})"

    @property
    def link(self) -> str:
        """Regular HTTPS link to the conversation through Slack.com."""
        return f"https://app.slack.com/client/{SlackApp().details['team_id']}/browse-user-groups/user_groups/{self.usergroup_id}"



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from django.conf import settings
from django.core.management.base import (
    BaseCommand,
    CommandError,
    CommandParser,
    OutputWrapper,
)

if TYPE_CHECKING:
    from celery.app.task import Task

    Task.__class_getitem__ = classmethod(lambda cls, *args, **kwargs: cls)  # type: ignore[attr-defined]


class Command(BaseCommand):
    help = "Generate the Slack App Manifest from your settings."
    requires_migrations_checks = False
    requires_system_checks = []

    def add_arguments(self, parser: CommandParser) -> None:
        output_format = parser.add_mutually_exclusive_group(required=True)
        output_format.add_argument(
            "--json",
            action="store_const",
            dest="output_format",
            const="json",
            help="Output the manifest as JSON.",
            default="json",
        )
        output_format.add_argument(
            "--yml",
            "--yaml",
            action="store_const",
            dest="output_format",
            const="yml",
            help="Output the manifest as YAML.",
        )
        # Add --public-url option for tunneling
        parser.add_argument(
            "--public-url",
            action="store",
            dest="public_url",
            help="The public URL to use for the manifest If not provided, BASE_URL will be used.",
            required=False,
        )

    def handle(self, *args: Any, **options: Any) -> None:
        output_format = options["output_format"]
        if output_format is None:
            raise CommandError("You must specify an output format (--json or --yml).")
        if output_format not in {"json", "yml"}:
            msg = f"Invalid output format: {output_format}. Must be either json or yml."
            raise CommandError(msg)

        base_url: str = settings.BASE_URL
        if not base_url:
            raise CommandError("You must specify a BASE_URL in your settings.")
        if not base_url.startswith("http"):
            base_url = f"https://{base_url}"
        app_display_name = settings.APP_DISPLAY_NAME

        public_base_url = (options.get("public_url") or base_url).rstrip("/")
        main_command: str = settings.SLACK_INCIDENT_COMMAND
        command_aliases: list[str] = settings.SLACK_INCIDENT_COMMAND_ALIASES
        if not main_command.startswith("/"):
            main_command = f"/{main_command}"
        if command_aliases:
            command_aliases = [
                f"/{alias}" if not alias.startswith("/") else alias
                for alias in command_aliases
            ]
            command_aliases = [
                alias for alias in command_aliases if alias != main_command
            ]

        manifest = get_manifest(
            app_display_name, main_command, command_aliases, public_base_url
        )
        output_manifest(manifest, output_format, self.stdout)


def get_manifest(
    app_display_name: str,
    main_command: str,
    command_aliases: list[str],
    public_base_url: str,
) -> dict[str, Any]:
    return {
        "display_information": {
            "name": app_display_name[:35],
            "description": "Incident Management Bot",
            "background_color": "#000000",
        },
        "features": {
            "app_home": {
                "home_tab_enabled": True,
                "messages_tab_enabled": False,
                "messages_tab_read_only_enabled": False,
            },
            "bot_user": {
                "display_name": app_display_name[:35],
                "always_online": True,
            },
            "shortcuts": [
                {
                    "name": "Open incident",
                    "type": "global",
                    "callback_id": "open_incident",
                    "description": "Open an incident and get help",
                }
            ],
            "slash_commands": [
                {
                    "command": command,
                    "url": f"{public_base_url}/api/v2/firefighter/slack/incident/",
                    "description": "Manage Incidents 🚨",
                    "usage_hint": "[open|update|close|status|help]",
                    "should_escape": False,
                }
                for command in [main_command, *command_aliases]
            ],
        },
        "oauth_config": {
            "scopes": {
                "bot": [
                    "bookmarks:read",
                    "bookmarks:write",
                    "channels:history",
                    "channels:join",
                    "channels:manage",
                    "channels:read",
                    "chat:write",
                    "chat:write.customize",
                    "chat:write.public",
                    "commands",
                    "groups:history",
                    "groups:read",
                    "groups:write",
                    "im:read",
                    "im:write",
                    "mpim:read",
                    "mpim:write",
                    "pins:read",
                    "pins:write",
                    "reactions:read",
                    "usergroups:read",
                    "usergroups:write",
                    "users.profile:read",
                    "users:read",
                    "users:read.email",
                    "metadata.message:read",
                    "team:read",
                    "incoming-webhook",
                ]
            }
        },
        "settings": {
            "event_subscriptions": {
                "request_url": f"{public_base_url}/api/v2/firefighter/slack/incident/",
                "bot_events": [
                    "app_home_opened",
                    "channel_archive",
                    "channel_deleted",
                    "channel_id_changed",
                    "channel_rename",
                    "channel_unarchive",
                    "group_archive",
                    "group_rename",
                    "group_unarchive",
                    "message.channels",
                    "message.groups",
                    "reaction_added",
                ],
            },
            "interactivity": {
                "is_enabled": True,
                "request_url": f"{public_base_url}/api/v2/firefighter/slack/incident/",
            },
            "org_deploy_enabled": False,
            "socket_mode_enabled": False,
            "token_rotation_enabled": False,
        },
    }


def output_manifest(
    manifest: dict[str, Any], output_format: str, stdout: OutputWrapper
) -> None:
    if output_format == "json":
        import json  # noqa: PLC0415

        stdout.write(json.dumps(manifest, indent=2, ensure_ascii=False))

    elif output_format == "yml":
        import yaml  # noqa: PLC0415

        stdout.write(yaml.dump(manifest, allow_unicode=True))



from __future__ import annotations

import logging
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
)

from django.conf import settings
from django.http import HttpRequest

from firefighter.incidents.enums import IncidentStatus

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence
    from datetime import datetime

    from django_htmx.middleware import HtmxDetails

    from firefighter.incidents.models.user import User


logger = logging.getLogger(__name__)

T = TypeVar("T", bound=dict[str, Any])
V = TypeVar("V")


def get_in(
    dictionary: dict[str, Any] | Any | None,
    keys: str | Sequence[str],
    default: Any | None = None,
) -> Any:
    """Get a value from arbitrarily nested dicts."""
    if dictionary is None:
        return None

    if isinstance(keys, str):
        keys = keys.split(".")
    if not keys:
        return dictionary
    if len(keys) == 1:
        return dictionary.get(keys[0], default)
    return get_in(dictionary.get(keys[0], {}), keys[1:], default=default)


def get_first_in(
    ulist: list[T],
    key: str | Sequence[str],
    matches: Iterable[str],
    default: V | None = None,
) -> T | V | None:
    """Returns the first element of a list of dicts, where the value of a key matches one in the provided iterable."""
    if not isinstance(ulist, list) or ulist is None:
        return default  # type: ignore[unreachable]
    return next(
        (e for e in ulist if get_in(e, key) in matches),
        default,
    )


def get_global_context(request: HttpRequest) -> dict[str, str | int | Any]:
    return {
        "APP_DISPLAY_NAME": settings.APP_DISPLAY_NAME,
        "FF_VERSION": settings.FF_VERSION,
        "PLAUSIBLE_SCRIPT_URL": settings.PLAUSIBLE_SCRIPT_URL,
        "PLAUSIBLE_DOMAIN": settings.PLAUSIBLE_DOMAIN,
        "IncidentStatus": IncidentStatus,
    }


def is_during_office_hours(dt: datetime) -> bool:
    """Check whether a datetime is during office hours. 9am-5pm, Mon-Fri.

    Args:
        dt (datetime): datetime with TZ info.
    """
    return (9 <= dt.hour <= 17) and (dt.weekday() < 5)


# Typing for custom HttpRequest classes
# Typing pattern recommended by django-stubs:
# https://github.com/typeddjango/django-stubs#how-can-i-create-a-httprequest-thats-guaranteed-to-have-an-authenticated-user
class HtmxHttpRequest(HttpRequest):
    htmx: HtmxDetails


class AuthenticatedHttpRequest(HttpRequest):
    user: User



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from django.conf import settings
from slack_sdk.models.blocks.blocks import SectionBlock

from firefighter.firefighter import utils
from firefighter.slack.slack_app import SlackApp
from firefighter.slack.slack_templating import (
    slack_block_help_commands,
    slack_block_help_description,
    slack_block_help_tip,
)
from firefighter.slack.views.modals import (
    modal_close,
    modal_dowgrade_workflow,
    modal_edit,
    modal_open,
    modal_postmortem,
    modal_send_sos,
    modal_status,
    modal_trigger_oncall,
    modal_update,
)

if TYPE_CHECKING:
    from slack_bolt.context.ack.ack import Ack
    from slack_bolt.context.respond.respond import Respond
APP_DISPLAY_NAME: str = settings.APP_DISPLAY_NAME
app = SlackApp()

logger = logging.getLogger(__name__)

SLACK_BUILTIN_COMMANDS = (
    "/archive",
    "/call",
    "/collapse",
    "/dm",
    "/expand",
    "/feed",
    "/invite",
    "/leave",
    "/msg",
    "/remind",
    "/remove",
    "/rename",
    "/search",
    "/shrug",
    "/status",
    "/topic",
)
"""List of all Slack built-in commands (https://slack.com/help/articles/201259356 checked August 2022)"""


def register_commands() -> None:
    """Register the command with its aliases.
    Commands are checked:
    - Fix commands that does not start with a slash, contain spaces, uppercase characters or are longer than 32 characters.
    - Ignore commands which names are built-in Slack commands.

    ⚠️ Don't forget to add the command and its aliases on your Slack App settings.
    """
    command_main: str = settings.SLACK_INCIDENT_COMMAND
    command_aliases: list[str] = settings.SLACK_INCIDENT_COMMAND_ALIASES

    commands = [command_main, *command_aliases]
    logger.debug(f"Registered commands: {commands}")
    for command in commands:
        if not command.startswith("/"):
            command = f"/{command}"  # noqa: PLW2901
            logger.warning(
                f"Command '{command}' does not start with a slash. We added one but please fix your configuration."
            )
        if " " in command:
            command = command.replace(" ", "")  # noqa: PLW2901
            logger.warning(
                f"Command '{command}' contained spaces. We removed them but please fix your configuration."
            )
        if len(command) > 32:
            command = command[:32]  # noqa: PLW2901
            logger.warning(
                f"Command '{command}' was longer than 32 characters. We truncated it but please fix your configuration."
            )
        if any(char.isupper() for char in command):
            command = command.lower()  # noqa: PLW2901
            logger.warning(
                f"Command '{command}' contained uppercase characters. We lower-cased them but please fix your configuration."
            )
        if command in SLACK_BUILTIN_COMMANDS:
            logger.warning(
                f"Command '{command}' is a built-in command, skipping registration. This command will not work."
            )
            continue
        app.command(command)(manage_incident)


def manage_incident(ack: Ack, respond: Respond, body: dict[str, Any]) -> None:
    logger.debug(body)
    command: str = utils.get_in(body, ["text"])

    if command is not None:
        command = command.lower().strip()

    if command == "open":
        modal_open.open_modal_aio(ack, body)
    elif command == "update":
        modal_update.open_modal_aio(ack, body)
    elif command == "edit":
        modal_edit.open_modal_aio(ack, body)
    elif command == "close":
        modal_close.open_modal_aio(ack=ack, body=body)
    elif command == "status":
        modal_status.open_modal_aio(ack=ack, body=body)
    elif command in {"oncall", "on-call"}:
        modal_trigger_oncall.open_modal_aio(ack=ack, body=body)
    elif command in {"postmortem", "post-mortem", "post mortem", "pm"}:
        modal_postmortem.open_modal_aio(ack=ack, body=body)
    elif command in {"sos", "sre", "sos-sre", "sos sre"}:
        modal_send_sos.open_modal_aio(ack=ack, body=body)
    elif command in {"downgrade", "convert"}:
        modal_dowgrade_workflow.open_modal_aio(ack=ack, body=body)
    elif command == "" or command is None:
        ack()
        send_help_message(respond=respond)
    else:
        ack()
        respond(text=f":x: Unknown command `{command}`!")
        send_help_message(respond=respond)


def send_help_message(respond: Respond) -> None:
    respond(
        text=f"{APP_DISPLAY_NAME} help",
        blocks=[
            SectionBlock(text=f"*{APP_DISPLAY_NAME} Commands Help*"),
            slack_block_help_description(),
            slack_block_help_commands(),
            slack_block_help_tip(),
        ],
    )


register_commands()



from __future__ import annotations

from firefighter.firefighter.settings.components.common import INSTALLED_APPS
from firefighter.firefighter.settings.settings_utils import config

ENABLE_JIRA: bool = config("ENABLE_JIRA", cast=bool, default=False)
"Enable the Jira app."

if ENABLE_JIRA:
    INSTALLED_APPS += ("firefighter.jira_app",)

    RAID_JIRA_API_USER: str = config("RAID_JIRA_API_USER")
    """The Jira API user to use."""
    RAID_JIRA_API_PASSWORD: str = config("RAID_JIRA_API_PASSWORD")
    """The Jira API password to use."""
    RAID_JIRA_API_URL: str = config("RAID_JIRA_API_URL")
    """The Jira API URL to use. If no protocol is defined, https will be used."""

    if not RAID_JIRA_API_URL.startswith("http"):
        RAID_JIRA_API_URL = f"https://{RAID_JIRA_API_URL}"
    RAID_JIRA_API_URL = RAID_JIRA_API_URL.rstrip("/")



"""MilestoneType model."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from django.db import models
from django.utils.translation import gettext_lazy as _
from django_stubs_ext.db.models import TypedModelMeta

if TYPE_CHECKING:
    import datetime


@dataclass
class MilestoneData:
    event_ts: datetime.datetime | None


def milestone_data_factory() -> MilestoneData:
    return MilestoneData(
        None,
    )


class MilestoneType(models.Model):
    """Represents a milestone type, also known as a key event time or event time."""

    name = models.CharField[str, str](
        max_length=50, help_text="Title or name (e.g. Recovered)"
    )
    event_type = models.SlugField(
        max_length=50,
        unique=True,
        help_text="Technical name, slugified version of the name (e.g. recovered)",
    )

    summary = models.CharField(
        max_length=250,
        help_text="Slack Markdown summary of the milestone (e.g. *Recovered*, when the impact is no longer apparent.)",
    )
    user_editable = models.BooleanField[bool, bool](
        help_text="Can be edited by users (else, system event)."
    )
    required = models.BooleanField[bool, bool](
        help_text="Is required for the incident to be considered as complete."
    )
    asked_for = models.BooleanField[bool, bool](
        default=True, help_text="Asked for in the form."
    )
    description = models.TextField(blank=True, null=True)

    class Meta(TypedModelMeta):
        verbose_name = _("Milestone Type")
        verbose_name_plural = _("Milestone Types")

    def __str__(self) -> str:
        return self.summary


@dataclass
class MilestoneTypeData:
    milestone_type: MilestoneType
    data: MilestoneData = field(default_factory=milestone_data_factory)



# Generated by Django 4.2.7 on 2023-12-04 13:24

import uuid

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    replaces = [
        ("pagerduty", "0001_squashed_0008"),
        ("pagerduty", "0009_alter_pagerdutyservice_pagerduty_id"),
    ]

    dependencies = [
        ("incidents", "0005_alter_incidentupdate_communication_lead"),
        ("incidents", "0006_alter_user_slack_id"),
        ("incidents", "0001_initial"),
        ("incidents", "0022_alter_user_name"),
    ]

    operations = [
        migrations.CreateModel(
            name="PagerDutyService",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("name", models.CharField(max_length=128)),
                (
                    "status",
                    models.CharField(
                        help_text="The current state of the Service. Valid statuses are:\n<ul>\n<li><code>active</code>: The service is enabled and has no open incidents. This is the only status a service can be created with.</li>\n<li><code>warning</code>: The service is enabled and has one or more acknowledged incidents.</li>\n<li><code>critical</code>: The service is enabled and has one or more triggered incidents.</li>\n<li><code>maintenance</code>: The service is under maintenance, no new incidents will be triggered during maintenance mode.</li>\n<li><code>disabled</code>: The service is disabled and will not have any new triggered incidents.</li>\n</ul>",
                        max_length=16,
                    ),
                ),
                (
                    "summary",
                    models.CharField(
                        help_text="A short-form, PD server-generated string that provides succinct, important information about an object suitable for primary labeling of an entity in a client. In many cases, this will be identical to name, though it is not intended to be an identifier. ",
                        max_length=256,
                    ),
                ),
                (
                    "api_url",
                    models.URLField(
                        help_text="The API show URL at which the object is accessible. Corresponds to PagerDuty API field `self`",
                        max_length=256,
                    ),
                ),
                (
                    "web_url",
                    models.URLField(
                        help_text="A URL at which the entity is uniquely displayed in the Web app. Corresponds to PagerDuty API field `html_url`",
                        max_length=256,
                    ),
                ),
                (
                    "pagerduty_id",
                    models.CharField(
                        db_index=True,
                        default=None,
                        help_text="PagerDuty ID for the service.",
                        max_length=64,
                        unique=True,
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                (
                    "ignore",
                    models.BooleanField(
                        default=False,
                        help_text="Ignore this service. Ignored services can't be triggered, and are hidden ins most places.",
                    ),
                ),
            ],
            options={
                "verbose_name": "PagerDuty service",
                "verbose_name_plural": "PagerDuty services",
            },
        ),
        migrations.CreateModel(
            name="PagerDutyIncident",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "title",
                    models.CharField(
                        help_text="A succinct description of the nature, symptoms, cause, or effect of the incident.",
                        max_length=256,
                    ),
                ),
                (
                    "summary",
                    models.CharField(
                        help_text="A short-form, server-generated string that provides succinct, important information about an object suitable for primary labeling of an entity in a client. In many cases, this will be identical to name, though it is not intended to be an identifier.",
                        max_length=256,
                    ),
                ),
                (
                    "status",
                    models.CharField(
                        help_text="The current status of the incident. Allowed values: triggered, acknowledged, resolved",
                        max_length=16,
                    ),
                ),
                (
                    "urgency",
                    models.CharField(
                        help_text="The urgency of the incident. Allowed values: high, low",
                        max_length=8,
                    ),
                ),
                (
                    "details",
                    models.CharField(
                        help_text="Additional incident details. Corresponds to PagerDuty API field `body.details`",
                        max_length=4000,
                    ),
                ),
                (
                    "api_url",
                    models.URLField(
                        help_text="The API show URL at which the object is accessible. Corresponds to PagerDuty API field `self`",
                        max_length=256,
                    ),
                ),
                (
                    "web_url",
                    models.URLField(
                        help_text="A URL at which the entity is uniquely displayed in the Web app. Corresponds to PagerDuty API field `html_url`",
                        max_length=256,
                    ),
                ),
                (
                    "incident_key",
                    models.CharField(
                        help_text="A string which identifies the incident. Sending subsequent requests referencing the same service and with the same incident_key will result in those requests being rejected if an open incident matches that incident_key.",
                        max_length=128,
                        unique=True,
                    ),
                ),
                (
                    "incident_number",
                    models.IntegerField(
                        help_text="The number of the incident. This is unique across your account."
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                (
                    "incident",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="pagerduty_incident_set",
                        to="incidents.incident",
                    ),
                ),
                (
                    "service",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to="pagerduty.pagerdutyservice",
                    ),
                ),
            ],
            options={
                "verbose_name": "PagerDuty incident",
                "verbose_name_plural": "PagerDuty incidents",
            },
        ),
        migrations.CreateModel(
            name="PagerDutyTeam",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("name", models.CharField(max_length=128, unique=True)),
                (
                    "description",
                    models.CharField(blank=True, max_length=1024, null=True),
                ),
                (
                    "pagerduty_id",
                    models.CharField(db_index=True, max_length=128, unique=True),
                ),
                (
                    "pagerduty_api_url",
                    models.URLField(blank=True, max_length=256, null=True),
                ),
                (
                    "pagerduty_url",
                    models.URLField(blank=True, max_length=256, null=True),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
            ],
            options={
                "verbose_name": "PagerDuty team",
                "verbose_name_plural": "PagerDuty teams",
            },
        ),
        migrations.CreateModel(
            name="PagerDutyEscalationPolicy",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "pagerduty_id",
                    models.CharField(db_index=True, max_length=128, unique=True),
                ),
                ("name", models.CharField(max_length=128, unique=True)),
                ("summary", models.CharField(blank=True, max_length=128, null=True)),
                (
                    "description",
                    models.CharField(blank=True, max_length=1024, null=True),
                ),
                (
                    "pagerduty_api_url",
                    models.URLField(blank=True, max_length=256, null=True),
                ),
                (
                    "pagerduty_url",
                    models.URLField(blank=True, max_length=256, null=True),
                ),
                (
                    "teams",
                    models.ManyToManyField(
                        blank=True,
                        related_name="pagerduty_escalation_policy_set",
                        to="pagerduty.pagerdutyteam",
                    ),
                ),
            ],
            options={
                "verbose_name": "PagerDuty escalation policy",
                "verbose_name_plural": "PagerDuty escalation policies",
            },
        ),
        migrations.CreateModel(
            name="PagerDutyUser",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "pagerduty_id",
                    models.CharField(db_index=True, max_length=128, unique=True),
                ),
                (
                    "pagerduty_api_url",
                    models.URLField(blank=True, max_length=256, null=True),
                ),
                (
                    "pagerduty_url",
                    models.URLField(blank=True, max_length=256, null=True),
                ),
                (
                    "phone_number",
                    models.CharField(blank=True, default="", max_length=128),
                ),
                (
                    "user",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="pagerduty_user",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "teams",
                    models.ManyToManyField(
                        blank=True,
                        related_name="pagerduty_user_set",
                        to="pagerduty.pagerdutyteam",
                    ),
                ),
                ("name", models.CharField(blank=True, max_length=256, null=True)),
            ],
            options={
                "verbose_name": "PagerDuty user",
                "verbose_name_plural": "PagerDuty users",
            },
        ),
        migrations.CreateModel(
            name="PagerDutySchedule",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "pagerduty_id",
                    models.CharField(db_index=True, max_length=128, unique=True),
                ),
                ("summary", models.CharField(blank=True, max_length=128, null=True)),
                (
                    "pagerduty_api_url",
                    models.URLField(blank=True, max_length=256, null=True),
                ),
                (
                    "pagerduty_url",
                    models.URLField(blank=True, max_length=256, null=True),
                ),
                (
                    "escalation_policies",
                    models.ManyToManyField(
                        blank=True,
                        related_name="pagerduty_schedule_set",
                        to="pagerduty.pagerdutyescalationpolicy",
                    ),
                ),
                (
                    "teams",
                    models.ManyToManyField(
                        blank=True,
                        related_name="pagerduty_schedule_set",
                        to="pagerduty.pagerdutyteam",
                    ),
                ),
                (
                    "users",
                    models.ManyToManyField(
                        blank=True,
                        related_name="pagerduty_schedule_set",
                        to="pagerduty.pagerdutyuser",
                    ),
                ),
            ],
            options={
                "verbose_name": "PagerDuty schedule",
                "verbose_name_plural": "PagerDuty schedules",
            },
        ),
        migrations.CreateModel(
            name="PagerDutyOncall",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("updated_at", models.DateTimeField(auto_now=True)),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                (
                    "escalation_level",
                    models.IntegerField(
                        help_text="The escalation level for the on-call."
                    ),
                ),
                (
                    "start",
                    models.DateTimeField(
                        blank=True,
                        help_text="The start of the on-call. If null, the on-call is a permanent user on-call.",
                        null=True,
                    ),
                ),
                (
                    "end",
                    models.DateTimeField(
                        blank=True,
                        help_text="The end of the on-call. If null, the user does not go off-call.",
                        null=True,
                    ),
                ),
                (
                    "escalation_policy",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="oncall_set",
                        to="pagerduty.pagerdutyescalationpolicy",
                    ),
                ),
                (
                    "pagerduty_user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="oncall_set",
                        to="pagerduty.pagerdutyuser",
                    ),
                ),
                (
                    "schedule",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="oncall_set",
                        to="pagerduty.pagerdutyschedule",
                    ),
                ),
            ],
            options={
                "verbose_name": "PagerDuty on-call",
                "verbose_name_plural": "PagerDuty on-calls",
            },
        ),
        migrations.AddField(
            model_name="pagerdutyservice",
            name="escalation_policy",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="pagerduty.pagerdutyescalationpolicy",
            ),
        ),
        migrations.AlterField(
            model_name="pagerdutyservice",
            name="pagerduty_id",
            field=models.CharField(
                db_index=True,
                help_text="PagerDuty ID for the service.",
                max_length=64,
                unique=True,
            ),
        ),
    ]



from __future__ import annotations

from django.urls import path

from firefighter.incidents.views import views
from firefighter.incidents.views.components.details import ComponentDetailView
from firefighter.incidents.views.components.list import ComponentsViewList
from firefighter.incidents.views.docs.metrics import MetricsView
from firefighter.incidents.views.docs.role_types import (
    RoleTypeDetailView,
    RoleTypeListView,
    RoleTypeRedirectView,
)
from firefighter.incidents.views.users.details import UserDetailView

app_name = "incidents"

urlpatterns = [
    # Incidents app views
    path("", views.DashboardView.as_view(), name="dashboard"),
    path(
        "incident/create/", views.IncidentCreateView.as_view(), name="incident-create"
    ),
    path("incident/", views.IncidentListView.as_view(), name="incident-list"),
    path(
        "incident/<int:incident_id>/",
        views.IncidentDetailView.as_view(),
        name="incident-detail",
    ),
    path(
        "incident/<int:incident_id>/update_key_events/",
        views.IncidentUpdateKeyEventsView.as_view(),
        name="incident-update-key-events",
    ),
    path(
        "statistics/",
        views.IncidentStatisticsView.as_view(),
        name="incident-statistics",
    ),
    path("component/", ComponentsViewList.as_view(), name="component-list"),
    path(
        "component/<uuid:component_id>/",
        ComponentDetailView.as_view(),
        name="component-detail",
    ),
    path(
        "user/<uuid:user_id>/",
        UserDetailView.as_view(),
        name="user-detail",
    ),
    path(
        "docs/metrics/",
        MetricsView.as_view(),
        name="docs-metrics",
    ),
    path("docs/role_types/", RoleTypeListView.as_view(), name="docs-role-type-list"),
    path(
        "docs/role_types/<int:pk>/",
        RoleTypeDetailView.as_view(),
        name="docs-role-type-detail",
    ),
    path(
        "docs/role_types/<slug:slug>/",
        RoleTypeRedirectView.as_view(),
        name="docs-role-type-detail-slug-redirect",
    ),
]



from __future__ import annotations

import logging

from django.apps import AppConfig

logger = logging.getLogger(__name__)


class RaidConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    label = "raid"
    name = "firefighter.raid"
    verbose_name = "RAID"

    def ready(self) -> None:
        import firefighter.raid.tasks
        import firefighter.raid.urls
        from firefighter.raid.signals import (
            incident_created,
            incident_updated,
        )
        from firefighter.raid.views.open_normal import (
            OpeningRaidCustomerModal,
            OpeningRaidDocumentationRequestModal,
            OpeningRaidFeatureRequestModal,
            OpeningRaidInternalModal,
            OpeningRaidSellerModal,
        )
        from firefighter.slack.views.modals.open import INCIDENT_TYPES

        INCIDENT_TYPES["normal"] = {
            "CUSTOMER": {
                "label": "Customer",
                "slack_form": OpeningRaidCustomerModal,
            },
            "SELLER": {
                "label": "Seller",
                "slack_form": OpeningRaidSellerModal,
            },
            "INTERNAL": {
                "label": "Internal",
                "slack_form": OpeningRaidInternalModal,
            },
            "DOCUMENTATION_REQUEST": {
                "label": "Documentation request",
                "slack_form": OpeningRaidDocumentationRequestModal,
            },
            "FEATURE_REQUEST": {
                "label": "Feature request",
                "slack_form": OpeningRaidFeatureRequestModal,
            },
        }

        return super().ready()



from __future__ import annotations

import random
import string
from typing import Any

from django.utils import timezone
from factory import LazyFunction, RelatedFactory, SubFactory
from factory.django import DjangoModelFactory
from factory.faker import Faker
from factory.fuzzy import FuzzyChoice
from faker import providers
from pyparsing import cast

from firefighter.incidents.factories import IncidentFactory, UserFactory
from firefighter.slack.models import SlackUser
from firefighter.slack.models.conversation import (
    Conversation,
    ConversationStatus,
    ConversationType,
)
from firefighter.slack.models.incident_channel import IncidentChannel
from firefighter.slack.models.message import Message


class SlackProvider(providers.BaseProvider):
    """Custom Faker provider for generating Slack IDs."""

    def slack_user_id(self) -> str:
        return cast(str, self.random_element("UW")) + "".join(  # type: ignore[redundant-cast]
            random.choices(  # noqa: S311
                string.ascii_uppercase + string.digits, k=self.random_int(8, 15)
            )
        )

    def slack_conversation_id(self) -> str:
        return cast(str, self.random_element("C")) + "".join(  # type: ignore[redundant-cast]
            random.choices(  # noqa: S311
                string.ascii_uppercase + string.digits, k=self.random_int(8, 15)
            )
        )


Faker.add_provider(SlackProvider)  # type: ignore[no-untyped-call]


class SlackUserFactory(DjangoModelFactory[SlackUser]):
    class Meta:
        model = SlackUser

    id = Faker("uuid4")  # type: ignore[no-untyped-call]
    slack_id = Faker("slack_user_id")  # type: ignore[no-untyped-call]
    user = SubFactory(UserFactory)  # type: ignore[no-untyped-call]
    username = Faker("user_name")  # type: ignore[no-untyped-call]
    image = Faker("image_url")  # type: ignore[no-untyped-call]


class SlackConversationFactory(DjangoModelFactory[Conversation]):
    class Meta:
        model = Conversation
        skip_postgeneration_save = True

    id = Faker("uuid4")  # type: ignore[no-untyped-call]
    name = Faker("name")  # type: ignore[no-untyped-call]
    channel_id = Faker("slack_conversation_id")  # type: ignore[no-untyped-call]
    _status = FuzzyChoice(ConversationStatus)  # type: ignore[no-untyped-call]
    _type = FuzzyChoice(ConversationType)  # type: ignore[no-untyped-call]
    is_shared = Faker("boolean")  # type: ignore[no-untyped-call]
    created_at = LazyFunction(timezone.now)  # type: ignore[no-untyped-call]
    updated_at = LazyFunction(timezone.now)  # type: ignore[no-untyped-call]
    members = RelatedFactory(UserFactory)  # type: ignore[no-untyped-call]
    tag = Faker("slug")  # type: ignore[no-untyped-call]


class IncidentChannelFactory(SlackConversationFactory):
    class Meta:
        model = IncidentChannel

    incident = SubFactory(IncidentFactory)  # type: ignore[no-untyped-call]


class MessageFactory(DjangoModelFactory):  # type: ignore[type-arg]
    class Meta:
        model = Message

    id = Faker("uuid4")  # type: ignore[no-untyped-call]
    conversation = SubFactory(SlackConversationFactory)  # type: ignore[no-untyped-call]
    ts = LazyFunction(timezone.now)  # type: ignore[no-untyped-call]
    user = SubFactory(SlackUserFactory)  # type: ignore[no-untyped-call]
    thread_ts = Faker("past_datetime", tzinfo=timezone.get_current_timezone())  # type: ignore[no-untyped-call]
    type = Faker("word")  # type: ignore[no-untyped-call]
    subtype = Faker("word")  # type: ignore[no-untyped-call]
    text = Faker("text", max_nb_chars=40000)  # type: ignore[no-untyped-call]
    blocks: dict[Any, Any] = {}  # JSONField can be represented as a dictionary
    incident = SubFactory(IncidentFactory)  # type: ignore[no-untyped-call]
    ff_type = Faker("random_element", elements=[Faker("word"), ""])  # type: ignore[no-untyped-call]



from __future__ import annotations

from firefighter.api.serializers import ComponentSerializer
from firefighter.api.views._base import ReadOnlyModelViewSet
from firefighter.incidents.models.component import Component


class ComponentViewSet(ReadOnlyModelViewSet[Component]):
    queryset = Component.objects.all().select_related("group")
    serializer_class = ComponentSerializer



import logging

from django.db import migrations, transaction

logger = logging.getLogger(__name__)

COMPONENT_SLACK_GROUPS = {
    "Controlling": ("S08UYJGJ9HB", "finance-controlling"),
    "Applications": ("S08V22ZJ6DQ", "corporate-it-applications"),
    "Gitlab": ("S090A6LQ0U8", "platform-gitlab"),
    "Seller Admin and Experience": ("S08UV261ZRU", "seller-seller-admin-and-experience"),
    "Revenue": ("S08V22ZPBBL", "finance-revenue"),
    "Catalog Exposition": ("S08VD2WE5ND", "catalog-catalog-exposition"),
    "Company reputation": ("S090A63V7FS", "marketing-communication-company-reputation"),
    "BO Catalog - Master Experience": ("S08VD2Y1DMF", "catalog-bo-catalog-master-experience"),
    "Customer service": ("S08UV2GB7UN", "operations-customer-service"),
    "Seller's invoices": ("S08VD38AFFB", "finance-sellers-invoices"),
    "Cart & funnel": ("S08V2N7PRN0", "marketplace-cart-funnel"),
    "Data Ingestion": ("S08UYJHEFEH", "data-data-ingestion"),
    "Refunds": ("S08UM14D3LP", "payment-operations-refunds"),
    "Data Analytics": ("S08V2NNDG9J", "data-data-analytics"),
    "Mobile Apps": ("S08UV20CL6S", "marketplace-mobile-apps"),
    "Navigation & Product discovery": ("S08UYHZFFD3", "marketplace-navigation-product-discovery"),
    "Helpcenter after sales": ("S08UU3SHQTV", "operations-helpcenter-after-sales"),
    "CDN": ("S08UM12RNTZ", "platform-cdn"),
    "Customer's invoices": ("S08UYJGELJH", "finance-customers-invoices"),
    "Returns": ("S08UYJHDNQM", "operations-returns"),
    "Inventory": ("S08V2NNLW3W", "operations-inventory"),
    "Loyalty and coupons": ("S08V230H1V0", "payment-operations-loyalty-and-coupons"),
    "Seller Services": ("S08UM0SJ66T", "seller-seller-services"),
    "Spinak": ("S08UYJGRTE1", "platform-spinak"),
    "Data Tools": ("S08V2NPFDPE", "data-data-tools"),
    "Product Management": ("S08UYJ5GU93", "catalog-product-management"),
    "Compromised laptop / server": ("S08UU3T6H51", "security-compromised-laptop-server"),
    "Catalog Performance": ("S08V22QDHJ6", "catalog-catalog-performance"),
    "Spartacux Foundations": ("S08UYHZH97X", "marketplace-spartacux-foundations"),
    "Stolen account(s) or IT materials": ("S08UU3TFJ3D", "security-stolen-accounts-or-it-materials"),
    "MM Fulfillment": ("S08VD371F0R", "operations-mm-fulfillment"),
    "HUB Integrators": ("S08UM0SSNNB", "seller-hub-integrators"),
    "Exploited vulnerability": ("S08V2NPD084", "security-exploited-vulnerability"),
    "Commercial Animation": ("S08V2N6UA3E", "marketplace-commercial-animation"),
    "Cloud Infrastructure": ("S08UYJJ01HB", "platform-cloud-infrastructure"),
    "Tracking": ("S08UYHZPSTX", "marketplace-tracking"),
    "Data Warehouse": ("S08UYJHRWQ5", "data-data-warehouse"),
    "Infra - System": ("S08UV2H5WDC", "corporate-it-infra-system"),
    "VAT": ("S090A6LGGV6", "finance-vat"),
    "Web performance": ("S090A63K124", "marketplace-web-performance"),
    "Payouts to seller": ("S090A6M61A4", "payment-operations-payouts-to-seller"),
    "Seller Catalog and Offer Management": ("S08UV261CHL", "seller-seller-catalog-and-offer-management"),
    "Other": ("S08V22Z8S58", "other-other"),
    "Network": ("S090A6M3B7S", "corporate-it-network"),
    "Bot management & rate limiting & WAF": ("S090A6M8740", "security-bot-management-rate-limiting-waf"),
    "Customer login & signup": ("S08V22H38AE", "marketplace-customer-login-signup"),
    "Delivery experience": ("S08V2NP1E5A", "operations-delivery-experience"),
    "Accounting": ("S08VD37CJBT", "finance-accounting"),
    "Traffic acquisition": ("S08UU3BNPPV", "marketing-communication-traffic-acquisition"),
    "Payment": ("S08UM13DTFH", "payment-operations-payment"),
    "Data leak": ("S08UYJHBJPP", "security-data-leak"),
    "Offer (Price & Stock)": ("S08V2NCSPJ8", "catalog-offer-price-stock"),
    "Product Structure": ("S08V2NCNFV2", "catalog-product-structure"),
    "Catalog Access": ("S090A69J064", "catalog-catalog-access"),
    "Order management": ("S08UU3TRGJF", "operations-order-management"),
}


def update_components_slack_groups(apps, _):
    try:
        Component = apps.get_model("incidents", "Component")
        UserGroup = apps.get_model("slack", "UserGroup")
    except LookupError:
        logger.exception("The 'slack' app is not installed. Skipping migration.")
        return

    with transaction.atomic():
        for component_name, (user_group_id, user_group_name) in COMPONENT_SLACK_GROUPS.items():
            try:
                # Vérifier/créer le SlackUserGroup
                user_group, _ = UserGroup.objects.get_or_create(
                    name=user_group_name
                )
                # Si le slack_group_id a changé, le mettre à jour
                if user_group.usergroup_id != user_group_id:
                    user_group.usergroup_id = user_group_id
                    user_group.handle = user_group_name
                    user_group.save()

                # Vérifier/créer le Component
                component = Component.objects.get(name=component_name)
                if not component:
                    logger.error(f"Component '{component_name}' n'existe pas.")
                    continue

                # Lier le Component au SlackUserGroup via la relation du UserGroup
                if not hasattr(user_group, "components"):
                    logger.error(f"UserGroup '{user_group_name}' n'a pas d'attribut 'components'.")
                    continue

                if not user_group.components.filter(id=component.id).exists():
                    user_group.components.add(component)
                    logger.info(f"Component '{component_name}' ajouté au SlackUserGroup '{user_group_name}'")
            except Exception:
                logger.exception(f"Failed to update component: '{component_name}'.")


def ensure_firefighters_user_group(apps, _):
    try:
        Component = apps.get_model("incidents", "Component")
        UserGroup = apps.get_model("slack", "UserGroup")
    except LookupError:
        logger.exception("The 'slack' app is not installed. Skipping Firefighters user group check.")
        return

    try:
        firefighters_group = UserGroup.objects.get(name="firefighters")
    except UserGroup.DoesNotExist:
        logger.exception("UserGroup 'firefighters' does not exist.")
        return

    for component in Component.objects.all():
        if not hasattr(firefighters_group, "components"):
            logger.error("UserGroup 'firefighters' n'a pas d'attribut 'components'.")
            break
        if not firefighters_group.components.filter(id=component.id).exists():
            firefighters_group.components.add(component)
            logger.info(f"Component '{component.name}' ajouté au SlackUserGroup 'firefighters'.")


def cleanup_unused_slack_user_groups(apps, _):
    try:
        UserGroup = apps.get_model("slack", "UserGroup")
    except LookupError:
        logger.exception("The 'slack' app is not installed. Skipping user groups cleanup.")
        return

    valid_usergroup_ids = {v[0] for v in COMPONENT_SLACK_GROUPS.values()}

    for user_group in UserGroup.objects.all():
        if (
            user_group.usergroup_id not in valid_usergroup_ids
            and "firefighter" not in user_group.name.lower()
            and hasattr(user_group, "components")
        ):
            components_to_remove = list(user_group.components.all())
            if components_to_remove:
                user_group.components.clear()
                logger.info(
                    f"Removed components {[c.name for c in components_to_remove]} from SlackUserGroup '{user_group.name}'"
                    )


def delete_unused_slack_user_groups(apps, _):
    try:
        UserGroup = apps.get_model("slack", "UserGroup")
    except LookupError:
        logger.exception("The 'slack' app is not installed. Skipping user groups deletion.")
        return

    for user_group in UserGroup.objects.all():
        if "firefighters" in user_group.name.lower():
            continue
        if hasattr(user_group, "components") and user_group.components.count() == 0:
            logger.info(f"Deleting SlackUserGroup '{user_group.name}' (ID: {user_group.usergroup_id}) as it has no components.")
            user_group.delete()


class Migration(migrations.Migration):

    dependencies = [
        ("incidents", "0013_add_missing_component"),
    ]

    operations = [
        migrations.RunPython(update_components_slack_groups),
        migrations.RunPython(ensure_firefighters_user_group),
        migrations.RunPython(cleanup_unused_slack_user_groups),
        migrations.RunPython(delete_unused_slack_user_groups),
    ]



from __future__ import annotations

from django.conf import settings
from django.contrib import admin

APP_DISPLAY_NAME: str = settings.APP_DISPLAY_NAME

admin.site.site_header = f"{APP_DISPLAY_NAME} Back-Office"
admin.site.site_title = f"{APP_DISPLAY_NAME} BO"
admin.site.index_title = f"{APP_DISPLAY_NAME} BO"

admin_custom = admin



"""From Django default errors views.
https://github.com/django/django/blob/0dd29209091280ccf34e07c9468746c396b7778e/django/views/defaults.py.
"""

from __future__ import annotations

import logging
from typing import Final
from urllib.parse import quote

from django.conf import settings
from django.http import (
    HttpRequest,
    HttpResponse,
    HttpResponseBadRequest,
    HttpResponseForbidden,
    HttpResponseNotFound,
    HttpResponseServerError,
    JsonResponse,
)
from django.template import Context, Engine, TemplateDoesNotExist, loader
from django.views.decorators.csrf import requires_csrf_token

logger = logging.getLogger(__name__)

JSON_CONTENT_TYPE: Final[str] = "application/json"


class JsonHttpResponseNotFound(JsonResponse, HttpResponseNotFound):
    pass


class JsonHttpResponseServerError(JsonResponse, HttpResponseServerError):
    pass


class JsonHttpResponseForbidden(JsonResponse, HttpResponseForbidden):
    pass


class JsonHttpResponseBadRequest(JsonResponse, HttpResponseBadRequest):
    pass


ERROR_404_TEMPLATE_NAME = ERROR_403_TEMPLATE_NAME = ERROR_400_TEMPLATE_NAME = (
    ERROR_500_TEMPLATE_NAME
) = "incidents/errors/base.html"

BACKUP_ERROR_PAGE_TEMPLATE = """
<!doctype html>
<html lang="en">
<head>
  <title>%(title)s</title>
</head>
<body>
  <h1>%(title)s</h1><p>%(details)s</p>
</body>
</html>
"""


# These views can be called when CsrfViewMiddleware.process_view() not run,
# therefore need @requires_csrf_token in case the template needs
# {% csrf_token %}.


@requires_csrf_token
def page_not_found(
    request: HttpRequest,
    exception: Exception,
    template_name: str = ERROR_404_TEMPLATE_NAME,
) -> HttpResponse:
    """Default 404 handler.

    Templates: :template:`404.html`
    Context:
        request_path
            The path of the requested URL (e.g., '/app/pages/bad_page/'). It's
            quoted to prevent a content injection attack.
        exception
            The message from the exception which triggered the 404 (if one was
            supplied), or the exception class name
    """
    if (
        request.headers.get("Accept", "") == JSON_CONTENT_TYPE
        or request.headers.get("Content-Type", "") == JSON_CONTENT_TYPE
    ):
        return JsonHttpResponseNotFound({
            "error": {"app_code": "NOT_FOUND", "message": "Page not found"}
        })
    exception_repr = exception.__class__.__name__
    # Try to get an "interesting" exception message, if any (and not the ugly
    # Resolver404 dictionary)
    try:
        message = exception.args[0] if exception.args else ""
    except (AttributeError, IndexError):
        logger.exception("Error in page_not_found")
    else:
        if isinstance(message, str):
            exception_repr = message
    context = {
        "request_path": quote(request.path),
        "exception": exception_repr,
        "page_title": "Page not found",
        "title": "Page not found",
        "code": "404",
        "message": "Please check the URL in the address bar and try again.",
    }

    try:
        template = loader.get_template(template_name)
        body = template.render(context, request)
    except TemplateDoesNotExist:
        if template_name != ERROR_404_TEMPLATE_NAME:
            # Reraise if it's a missing custom template.
            raise
        # Render template (even though there are no substitutions) to allow
        # inspecting the context in tests.
        template_backup = Engine().from_string(
            BACKUP_ERROR_PAGE_TEMPLATE
            % {
                "title": "Not Found",
                "details": "The requested resource was not found on this server.",
            },
        )

        body = template_backup.render(Context(context))
    return HttpResponseNotFound(body)


@requires_csrf_token
def server_error(
    request: HttpRequest, template_name: str = ERROR_500_TEMPLATE_NAME
) -> HttpResponse:
    """500 error handler.

    Templates: :template:`500.html`
    Context: None
    """
    context = {
        "page_title": "Server error",
        "title": "Server error",
        "code": "500",
        "message": "Please try again later. This issue has been logged, but feel free to raise the issue.",
        "APP_DISPLAY_NAME": settings.APP_DISPLAY_NAME,
    }
    if (
        request.headers.get("Accept", "") == JSON_CONTENT_TYPE
        or request.headers.get("Content-Type", "") == JSON_CONTENT_TYPE
    ):
        return JsonHttpResponseServerError({
            "error": {"app_code": "SERVER_ERROR", "message": "Server error"}
        })
    try:
        template = loader.get_template(template_name)
    except TemplateDoesNotExist:
        if template_name != ERROR_500_TEMPLATE_NAME:
            # Reraise if it's a missing custom template.
            raise
        return HttpResponseServerError(
            BACKUP_ERROR_PAGE_TEMPLATE % {"title": "Server Error (500)", "details": ""},
        )
    return HttpResponseServerError(template.render(context))


# pylint: disable=unused-argument
@requires_csrf_token
def bad_request(
    request: HttpRequest,
    exception: Exception,
    template_name: str = ERROR_400_TEMPLATE_NAME,
) -> HttpResponse:
    """400 error handler.

    Templates: :template:`400.html`
    Context: None
    """
    if (
        request.headers.get("Accept", "") == JSON_CONTENT_TYPE
        or request.headers.get("Content-Type", "") == JSON_CONTENT_TYPE
    ):
        return JsonHttpResponseBadRequest({
            "error": {"app_code": "BAD_REQUEST", "message": "Bad request"}
        })
    context = {
        "request_path": quote(request.path),
        "page_title": "Bad request",
        "APP_DISPLAY_NAME": settings.APP_DISPLAY_NAME,
        "title": "Bad request",
        "code": "400",
        "message": "Please check the URL in the address bar and try again.",
    }
    try:
        template = loader.get_template(template_name)
    except TemplateDoesNotExist:
        if template_name != ERROR_400_TEMPLATE_NAME:
            # Reraise if it's a missing custom template.
            raise
        return HttpResponseBadRequest(
            BACKUP_ERROR_PAGE_TEMPLATE % {"title": "Bad Request (400)", "details": ""},
        )
    # No exception content is passed to the template, to not disclose any
    # sensitive information.
    return HttpResponseBadRequest(template.render(context))


@requires_csrf_token
def permission_denied(
    request: HttpRequest,
    exception: Exception,
    template_name: str = ERROR_403_TEMPLATE_NAME,
) -> HttpResponse:
    """Permission denied (403) handler.

    Templates: :template:`403.html`
    Context:
        exception
            The message from the exception which triggered the 403 (if one was
            supplied).

    If the template does not exist, an HTTP 403 response containing the text
    "403 Forbidden" (as per RFC 7231) will be returned.
    """
    if (
        request.headers.get("Accept", "") == JSON_CONTENT_TYPE
        or request.headers.get("Content-Type", "") == JSON_CONTENT_TYPE
    ):
        return JsonHttpResponseForbidden({
            "error": {"app_code": "FORBIDDEN", "message": "Forbidden"}
        })
    context = {
        "request_path": quote(request.path),
        "exception": str(exception),
        "page_title": "Forbidden",
        "title": "Forbidden",
        "code": "403",
        "message": "You do not have permission to access this page.",
    }
    try:
        template = loader.get_template(template_name)
    except TemplateDoesNotExist:
        if template_name != ERROR_403_TEMPLATE_NAME:
            # Reraise if it's a missing custom template.
            raise
        return HttpResponseForbidden(
            BACKUP_ERROR_PAGE_TEMPLATE % {"title": "403 Forbidden", "details": ""},
        )
    return HttpResponseForbidden(template.render(request=request, context=context))



from __future__ import annotations

from typing import Any

from django.core.management.base import BaseCommand

from firefighter.confluence.tasks import sync_runbooks


class Command(BaseCommand):
    help = "Sync runbooks from Confluence. This is a wrapper for the [firefighter.confluence.tasks.sync_runbooks][] task."

    def handle(self, *args: Any, **options: Any) -> Any:
        return sync_runbooks()



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Never

from django.conf import settings
from django.dispatch.dispatcher import receiver

from firefighter.slack.messages.slack_messages import (
    SlackMessageIncidentComProcess,
    SlackMessageIncidentRolesUpdated,
)
from firefighter.slack.signals import incident_channel_done

if TYPE_CHECKING:
    from firefighter.incidents.models.incident import Incident
    from firefighter.slack.models.incident_channel import IncidentChannel

logger = logging.getLogger(__name__)
SLACK_APP_EMOJI: str = settings.SLACK_APP_EMOJI


# noinspection PyUnusedLocal
@receiver(signal=incident_channel_done)
def incident_channel_done_bookmarks(
    incident: Incident, channel: IncidentChannel, **kwargs: Any
) -> None:
    channel.add_bookmark(
        title="Incident Page",
        link=incident.status_page_url
        + "?utm_medium=FireFighter+Slack&utm_source=Slack+Bookmark&utm_campaign=Bookmark+In+Channel",
        emoji=SLACK_APP_EMOJI,
    )


# noinspection PyUnusedLocal
@receiver(incident_channel_done)
def sev1_process_reminder(
    incident: Incident, channel: IncidentChannel, **kwargs: Never
) -> None:
    if incident.priority.value == 1 and incident.environment.value == "PRD":
        channel.send_message_and_save(SlackMessageIncidentComProcess(incident))


@receiver(incident_channel_done)
def send_roles_message_in_conversation(
    incident: Incident, channel: IncidentChannel, **kwargs: Any
) -> None:
    update_roles_message = SlackMessageIncidentRolesUpdated(
        incident=incident,
        incident_update=None,
        first_update=True,
    )
    incident.conversation.send_message_and_save(update_roles_message)



from __future__ import annotations

from typing import TYPE_CHECKING

from django.db import models
from django_stubs_ext.db.models import TypedModelMeta

if TYPE_CHECKING:
    from firefighter.incidents.models.incident import Incident  # noqa: F401
    from firefighter.incidents.models.incident_cost_type import (  # noqa: F401
        IncidentCostType,
    )


class IncidentCost(models.Model):
    """Incident Cost is inspired from dispatch."""

    id = models.BigAutoField(primary_key=True)

    amount = models.DecimalField(
        max_digits=10,
        decimal_places=2,
        null=True,
        blank=True,
        help_text="Amount in EUR",
    )
    details = models.TextField(null=True, blank=True)

    cost_type = models.ForeignKey(
        "IncidentCostType", on_delete=models.CASCADE, related_name="incident_cost_set"
    )
    incident = models.ForeignKey(
        "Incident", on_delete=models.CASCADE, related_name="incident_cost_set"
    )

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta(TypedModelMeta):
        verbose_name = "Incident cost"
        verbose_name_plural = "Incident costs"

    def __str__(self) -> str:
        return f"{self.cost_type.name} - {self.amount}"

    @property
    def currency(self) -> str:
        return "EUR"
